# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.utils import Const
from msprobe.core.dump.service import BaseService
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import get_rank_if_initialized
from msprobe.pytorch.dump.module_dump.module_processor import ModuleProcessor
from msprobe.pytorch.dump.api_dump.api_register import (
    get_api_register,
    ApiTemplate,
    redirect_wait,
    reset_dist_collect_func
)
from msprobe.pytorch.dump.api_dump.hook_module import HOOKModule
from msprobe.pytorch.dump.api_dump.pt_hook_manager import PytorchHookManager
from msprobe.pytorch.dump.api_dump.register_optimizer_hook import register_optimizer_hook
from msprobe.pytorch.dump.api_dump.script_wrapper import wrap_script_func, preprocess_func


class PytorchService(BaseService):
    @property
    def _get_framework_type(self):
        return Const.PT_FRAMEWORK

    @staticmethod
    def _get_current_rank():
        return get_rank_if_initialized()

    def reset_status(self):
        self._reset_status()

    def _init_specific_components(self):
        self.logger = logger
        self.api_register = get_api_register()
        self.module_processor = ModuleProcessor(self.data_collector.scope)
        self.hook_manager = PytorchHookManager(self.data_collector, self.config)
        self.api_template = ApiTemplate

    def _register_hook(self):
        if self._is_mix_level:
            register_optimizer_hook(self.data_collector)

    def _register_api_hook(self):
        preprocess_func()
        super()._register_api_hook()
        wrap_script_func()
        redirect_wait()

    def _register_module_hook(self):
        ModuleProcessor.enable_module_dump = True
        self.module_processor.register_module_hook(self.model, self.build_hook)
        self.logger.info(f"The module {self.config.task} hook function is successfully mounted to the model.")

    def _reset_status(self):
        super()._reset_status()
        ModuleProcessor.reset_module_stats()
        HOOKModule.reset_module_stats()
        reset_dist_collect_func()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the compare_npu_vs_npu function.
"""
import argparse
import multiprocessing
import os
import csv
import time
import numpy as np

from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check, log
from msprobe.msaccucmp.pytorch_cmp import pytorch_dump_data as pytorch_dump
from msprobe.msaccucmp.algorithm_manager.algorithm_manager import AlgorithmManager
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.vector_cmp.fusion_manager import compare_result
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.utils import sanitize_csv_value


class PytorchComparison:
    """
    The class for comparing PyTorch dump data.
    """
    OP_HEADER = ["NPUDump", "NPUDumpPath", "GroundTruthPath", "DataType"]

    def __init__(self: any, args: argparse.Namespace) -> None:
        self.compare_data = pytorch_dump.CompareData(
            os.path.realpath(args.my_dump_path),
            os.path.realpath(args.golden_dump_path))
        file_name = 'result_%s.csv' \
                    % time.strftime("%Y%m%d%H%M%S",
                                    time.localtime(time.time()))
        ret = path_check.check_output_path_valid(args.output_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        if os.path.islink(os.path.abspath(args.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.join(
            os.path.realpath(args.output_path), file_name)
        self.algorithm_manager = AlgorithmManager(args.custom_script_path,
                                                  args.algorithm,
                                                  args.algorithm_options)
        self.is_open_advisor = args.advisor
        self.filter_flag = args.post_process

    @staticmethod
    def _save_numpy_data(file_path: str, data: any) -> None:
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path), mode=0o700)
        np.save(file_path, data)

    @staticmethod
    def _get_item_location(row: list) -> list:
        cos_index = 0
        my_dump_index = 0
        golden_index = 0
        for (index, item) in enumerate(row):
            if item == "CosineSimilarity":
                cos_index = index
            elif item == "MyDumpDataPath":
                my_dump_index = index
            elif item == "GoldenDumpDataPath":
                golden_index = index
        return [cos_index, my_dump_index, golden_index]

    def record_not_matched(self: any) -> None:
        """
        Find all datasets that do not match my dump in the golden dump and record
        them in the result file.
        """
        not_matched_result = []
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager)
        for order, op_name, golden_dataset, message in \
                self.compare_data.get_not_matched_golden_datasets():
            not_matched_result += fusion_op_result.get_pytorch_result(
                compare_result.PytorchOpInfo(order, op_name, ConstManager.NAN, golden_dataset), None, [message])
        self._save_cmp_result(not_matched_result)

    def compare(self: any) -> int:
        """
        Compare for pytorch dump data.
        """
        log.print_info_log("start parse dump file!")
        # Write the result header
        if not self._write_header_to_file():
            return CompareError.MSACCUCMP_OPEN_FILE_ERROR

        # parse and compare
        try:
            ret = self.compare_data.parse_dump_file()
        except CompareError as error:
            return error.code
        finally:
            self.compare_data.close_file()

        # compare dump file
        if ret == CompareError.MSACCUCMP_NONE_ERROR:
            return self._compare_net()
        return ret

    def check_arguments_valid(self: any, args: argparse.Namespace) -> None:
        """
        Check arguments valid, if invalid, throw exception
        """
        check_pass = True
        if len(args.fusion_rule_file) != 0 \
                or len(args.quant_fusion_rule_file) != 0 \
                or len(args.custom_script_path) != 0:
            check_pass = False

        if args.op_name is not None \
                or args.dump_version != ConstManager.BINARY_DUMP_TYPE \
                or args.mapping:
            check_pass = False

        if not check_pass:
            log.print_error_log('The argument [-f|-q|-c|-op|-map|-v] is not supported'
                                ' in pytorch precision comparison scenarios.')
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

        path_type = path_check.PathType.File
        ret = path_check.check_output_path_valid(self.output_path, False, path_type)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

    def _write_header_to_file(self: any) -> bool:
        """
        Write the result file header into the file.
        """
        try:
            with os.fdopen(os.open(self.output_path, ConstManager.WRITE_FLAGS,
                                   ConstManager.WRITE_MODES), 'a+', newline='') as output_file:
                compare_result_header = compare_result.get_result_title(self.algorithm_manager, self.OP_HEADER)
                writer = csv.writer(output_file)
                if writer is not None:
                    sanitized_compare_result_header = [sanitize_csv_value(cell) for cell in compare_result_header]
                    writer.writerow(sanitized_compare_result_header)
            return True
        except IOError as io_error:
            log.print_open_file_error(self.output_path, io_error)
            return False
        finally:
            pass

    def _save_cmp_result_exec(self: any, result: list) -> None:
        with os.fdopen(os.open(self.output_path, ConstManager.WRITE_FLAGS,
                               ConstManager.WRITE_MODES), 'a+', newline='') as output_file:
            writer = csv.writer(output_file)
            for item in result:
                sanitized_row = [sanitize_csv_value(cell) for cell in item]
                writer.writerow(sanitized_row)

    def _save_cmp_result(self: any, result: list, lock: any = None) -> None:
        """
        Write the compare result to result file.
        """
        if lock is not None:
            lock.acquire()
        try:
            self._save_cmp_result_exec(result)
        except IOError as io_error:
            log.print_open_file_error(self.output_path, io_error)
        finally:
            if lock is not None:
                lock.release()

    def _get_compare_dump_data(self: any, op_name: str, my_dump_dataset: str, golden_dataset: str) -> any:
        """
        Get compare data and check it
        """
        my_dump_data, golden_dump_data, _ = self.compare_data.get_dump_data(
            my_dump_dataset, golden_dataset)
        if my_dump_data.size != golden_dump_data.size:
            message = log.print_cannot_compare_warning(
                op_name, '(%d)' % my_dump_data.size, '(%d)' % golden_dump_data.size)
            raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR, message)
        return my_dump_data.flatten(), golden_dump_data.flatten(), my_dump_data.shape

    def _do_compare_tensor(self: any, op_name: str, my_dump_dataset: str, golden_dataset: str) -> any:
        my_dump_data, golden_dump_data, shape = self._get_compare_dump_data(
            op_name, my_dump_dataset, golden_dataset)
        my_dump_data_type = str(my_dump_data.dtype)
        algorithm_result, fail_reason = self.algorithm_manager.compare(
            my_dump_data, golden_dump_data,
            {'my_output_dump_file': my_dump_dataset, 'ground_truth_dump_file': golden_dataset,
             'shape_type': utils.get_shape_type(shape)})
        return [my_dump_data_type, shape], algorithm_result, fail_reason

    def _compare_tensor(self: any, order: int, op_name: str, ext_opname: str, my_dump_dataset: str) -> any:
        """
        compare one tensor
        :order: order
        :name: name
        :ext_opname: extend_opname
        :my_dump_dataset: dataset path for compare
        """
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager)
        match, golden_dataset, fail_reason = self.compare_data.get_golden_dataset(ext_opname, my_dump_dataset)

        if not match:
            return match, fusion_op_result.get_pytorch_result(
                compare_result.PytorchOpInfo(order, op_name, my_dump_dataset, ConstManager.NAN), None, [fail_reason])
        op_info = compare_result.PytorchOpInfo(order, op_name, my_dump_dataset, golden_dataset)
        try:
            return match, self._do_compare_and_get_result([my_dump_dataset, golden_dataset],
                                                          fusion_op_result, op_info)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                CompareError) as err:
            error_msg = [err.message] if isinstance(err, CompareError) else [str(err)]
            return match, fusion_op_result.get_pytorch_result(op_info, None, error_msg)
        finally:
            pass

    def _do_compare_and_get_result(self: any, dataset_list: list,
                                   fusion_op_result: compare_result.FusionOpComResult,
                                   op_info: compare_result.PytorchOpInfo) -> list:
        type_shape_list, algorithm_result, fail_reason = self._do_compare_tensor(op_info.op_name, dataset_list[0],
                                                                                 dataset_list[1])
        tensor_info = {
            "tensor_id": None,
            "shape": type_shape_list[1],
            "my_output_dtype": type_shape_list[0]
        }
        return fusion_op_result.get_pytorch_result(
                op_info,
                [compare_result.TensorResult(tensor_info, [algorithm_result, ''], fail_reason, False)],
                fail_reason)

    def _compare_one_op(self: any, order: int, ext_opname: str, lock: any) -> None:
        """
        Compare the dump data corresponding to ext_opname.
        """
        my_dump_datasets = self.compare_data.get_my_dump_datasets(ext_opname)
        op_name = self.compare_data.get_original_opname(ext_opname)
        log.print_info_log('[{0}] Start to compare "/{0}/{1}".'.format(op_name, order))
        if not my_dump_datasets:
            log.print_warn_log("[{}] has no comparable data.".format(op_name))
            return

        not_match = []
        op_cmp_result = []
        for my_dump_dataset in my_dump_datasets:
            match, tensor_result = self._compare_tensor(order, op_name, ext_opname, my_dump_dataset)
            not_match.append(not match)
            op_cmp_result += tensor_result
        if all(not_match):
            op_cmp_result.clear()
            fail_reason = 'No data match with op:{} in golden dump data.'.format(op_name)
            op_cmp_result = compare_result.FusionOpComResult(self.algorithm_manager).get_pytorch_result(
                compare_result.PytorchOpInfo(order, op_name, ConstManager.NAN, ConstManager.NAN), None, [fail_reason])
        self._save_cmp_result(op_cmp_result, lock)

    def _compare_in_one_process(self: any, order_group: list, lock: any) -> None:
        """
        Compare all the ops in the order_group.
        """
        # the parallel reading of the h5py file needs to be \
        # opened separately in the process.
        self.compare_data.open_file('r')
        try:
            for order in order_group:
                for ext_opname in self.compare_data.get_ext_opname_by_order(order):
                    self._compare_one_op(order, ext_opname, lock)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                CompareError) as ex:
            log.print_warn_log("Failed to compare with except: " + str(ex))
        finally:
            self.compare_data.close_file()

    def _compare_by_multi_process(self: any) -> None:
        """
        Compare for all the pytorch op by multi process
        """
        process_num = int((multiprocessing.cpu_count() + 1) / 2)
        all_orders = self.compare_data.get_all_orders()
        orders_groups = (all_orders[i::process_num] for i in range(process_num))

        # start multiple processes for parallel comparison.
        pool = multiprocessing.Pool(process_num)
        lock = multiprocessing.Manager().RLock()
        for order_group in orders_groups:
            pool.apply_async(self._compare_in_one_process, args=(order_group, lock))

        pool.close()
        pool.join()

    def _filter_one_line(self: any, result_path: str, row: list, csv_writer: any, position: list) -> None:
        cos_index = position[0]
        my_dump_index = position[1]
        golden_index = position[2]
        if row[cos_index] == ConstManager.NAN or float(row[cos_index]) <= 0.95:
            sanitized_row = [sanitize_csv_value(cell) for cell in row]
            csv_writer.writerow(sanitized_row)
            if row[my_dump_index] != ConstManager.NAN:
                tensor_data = self.compare_data.my_dump.get_dump_data(row[my_dump_index])
                dump_data_path = os.path.join(result_path, "my_dump" + row[my_dump_index])
                self._save_numpy_data(dump_data_path, tensor_data)

            if row[golden_index] != ConstManager.NAN:
                tensor_data = self.compare_data.golden_dump.get_dump_data(row[golden_index])
                dump_data_path = os.path.join(result_path, "gold_dump" + row[golden_index])
                self._save_numpy_data(dump_data_path, tensor_data)

    def _filter_result_process(self: any,
                               result_file_handle: any,
                               filtered_file_handle: any,
                               filtered_result_path: str) -> None:
        csv_reader = csv.reader(result_file_handle)
        csv_writer = csv.writer(filtered_file_handle)
        line_num = 0
        location = []
        for row in csv_reader:
            if line_num == 0:
                location = self._get_item_location(row)
                sanitized_row = [sanitize_csv_value(cell) for cell in row]
                csv_writer.writerow(sanitized_row)
            else:
                self._filter_one_line(filtered_result_path, row, csv_writer, location)
            line_num = line_num + 1

    def _filter_compare_result(self: any) -> None:
        result_filename_base, _ = os.path.splitext(os.path.basename(self.output_path))
        filtered_result_folder_name = "{}_{}".format(result_filename_base, "filtered")
        filtered_result_folder_path = os.path.join(os.path.dirname(self.output_path), filtered_result_folder_name)
        if not os.path.exists(filtered_result_folder_path):
            os.makedirs(filtered_result_folder_path, mode=0o700)

        filtered_result_file = os.path.join(filtered_result_folder_path, "filtered_result.csv")
        self.compare_data.open_file('r')
        try:
            with open(self.output_path, 'r') as result_file:
                with os.fdopen(os.open(filtered_result_file, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                               'w+') as filtered_file:
                    self._filter_result_process(result_file, filtered_file, filtered_result_folder_path)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            log.print_error_log('Failed to filtering pytorch compare result!, {}'.format(str(error)))
        finally:
            self.compare_data.close_file()

    def _compare_net(self: any) -> int:
        """
        compare vector for all fusion ops, and write result to file.
        """
        log.print_info_log("pytorch precision compare start!")
        self._compare_by_multi_process()
        self.record_not_matched()
        utils.sort_result_file_by_index(self.output_path)

        if os.path.exists(self.output_path):
            log.print_write_result_info('comparison result', self.output_path)
            if self.filter_flag:
                self._filter_compare_result()
            if self.is_open_advisor:
                self._do_advisor()
            return CompareError.MSACCUCMP_NONE_ERROR
        return CompareError.MSACCUCMP_UNKNOWN_ERROR

    def _do_advisor(self):
        try:
            from msprobe.msaccucmp.advisor.compare_advisor import CompareAdvisor
        except ImportError as import_error:
            log.print_warn_log("Unable to import module: %s." % str(import_error))
            log.print_warn_log("Skip compare results Analysis.")
        else:
            out_path = os.path.dirname(self.output_path)
            compare_advisor = CompareAdvisor(self.output_path, [], out_path)
            advisor_result = compare_advisor.advisor()
            message_list = advisor_result.print_advisor_log()
            advisor_result.gen_summary_file(out_path, message_list)




# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch
from torch.autograd.functional import jacobian
from msprobe.pytorch.common.log import logger


@torch.no_grad()
def square_sum(x: torch.tensor):
    return (x * x).sum()


@torch.no_grad()
def get_min(x: torch.tensor):
    return torch.min(x)


@torch.no_grad()
def get_mean(x: torch.tensor):
    return torch.mean(x.to(torch.float64))


@torch.no_grad()
def get_norm(x: torch.tensor):
    return torch.norm(x.to(torch.float64), p=2)


@torch.no_grad()
def get_max(x: torch.tensor):
    return torch.max(x)


@torch.no_grad()
def get_zeros(x: torch.tensor, eps: float):
    if x.numel() == 0:
        return torch.tensor(float('nan'))
    return torch.sum(torch.abs(x) < eps) / x.numel()


@torch.no_grad()
def get_sign_matches(x: torch.tensor, y: torch.tensor):
    if y.numel() == 0:
        return torch.tensor(1.)
    xs = x.sign()
    ys = y.sign()

    try:
        same_direction_ratio = ((xs * ys).sum() / ys.numel() + 1) / 2
    except RuntimeError as e:
        logger.info(f"RuntimeError: {e}")
        same_direction_ratio = torch.tensor(0.)
    return same_direction_ratio


@torch.no_grad()
def eff_rank(param: torch.tensor, threshold=1e-10):
    U, S, Vh = torch.linalg.svd(param.float())
    rank = torch.sum(S > threshold)
    return rank


# modular neural tangent kernel
@torch.no_grad()
def mNTK(module: torch.nn.Module, x: torch.tensor):
    J_theta_l = jacobian(module, x)
    mntk = torch.matmul(J_theta_l, J_theta_l.t())
    return mntk


@torch.no_grad()
def power_iteration(a, num_iterations):
    b = torch.randn(a.size(1), 1)
    for _ in range(num_iterations):
        b = torch.matmul(a, b)
        b_norm = torch.norm(b)
        b = b / b_norm if b_norm != 0 else 0
    eigval = torch.matmul(torch.matmul(b.t(), a), b)
    return eigval


@torch.no_grad()
def lambda_max_subsample(module: torch.nn.Module, x: torch.tensor, num_iterations=100, subsample_size=None):
    mntk = mNTK(module, x)
    if subsample_size is None:
        subsample_size = min(mntk.size(0), mntk.size(1))
    idx = torch.randperm(mntk.size(0))[:subsample_size]
    subsampled = mntk[idx, :]
    subsampled = subsampled[:, idx]
    eigval = power_iteration(subsampled, num_iterations)
    return eigval


@torch.no_grad()
def cal_histc(tensor_cal, bins_total, min_val, max_val):
    return torch.histc(tensor_cal, bins=bins_total, min=min_val, max=max_val)


@torch.no_grad()
def get_nans(t):
    return torch.isnan(t).sum()


def check_tensor_dim(tensor, n):
    """检查张量维度是否大于n
    """
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(
            f"Input must be a PyTorch tensor. Got {type(tensor)} instead. "
            f"Consider using torch.tensor() for conversion."
        )

    if tensor.dim() < n:
        raise ValueError(
            f"Tensor must have at least {n} dimensions. "
            f"Got shape: {tuple(tensor.shape)} with {tensor.dim()} dims."
        )


@torch.no_grad()
def max_eigenvalue(input_tensor: torch.Tensor, num_iterations=3):
    input_tensor = input_tensor.float()
    try:
        check_tensor_dim(input_tensor, 2)
    except (TypeError, ValueError) as e:
        logger.warning(f"Calculate max eigenvalue failed: {e}")
        return torch.tensor(0)
    in_features = input_tensor.shape[1]
    u_tensor = torch.randn(in_features).to(input_tensor.device)
    u_norm = u_tensor.norm()
    if u_norm.item() == 0:
        return torch.tensor(0)
    u_tensor = u_tensor / u_tensor.norm()
    input_seq = torch.matmul(input_tensor.T, input_tensor)
    for _ in range(num_iterations):
        v_tensor = torch.matmul(input_seq, u_tensor)
        spectral_norm = torch.matmul(v_tensor.T, u_tensor)
        v_norm = v_tensor.norm()
        if v_norm > 0:
            u_tensor = v_tensor / v_norm
        else:
            spectral_norm = torch.tensor(0)
            break
    return spectral_norm.sqrt()


@torch.no_grad()
def cal_entropy(qk_tensor, mask=None):
    try:
        check_tensor_dim(qk_tensor, 2)
    except (TypeError, ValueError) as e:
        logger.warning(f"Calculate max eigenvalue failed: {e}")
        return torch.tensor(0), torch.tensor(0)
    if mask is None:
        mask = torch.tril(torch.ones(qk_tensor.shape[1], qk_tensor.shape[1])).to(
            qk_tensor.device)
    qk_tensor = qk_tensor - torch.amax(qk_tensor, dim=1, keepdim=True)
    qk_tensor = qk_tensor.masked_fill(mask == 0, float('-inf'))
    softmax_qkt = torch.nn.functional.softmax(qk_tensor.float(), dim=1)
    # softmax取QK矩阵最大值
    softmax_max = torch.mean(torch.amax(softmax_qkt, dim=1))
    entropy = torch.mean(-torch.nansum(softmax_qkt *
                         torch.log(softmax_qkt), dim=1))
    return entropy, softmax_max


@torch.no_grad()
def cal_qkt(q_h, k_h, order="s,b,h,d"):
    # q_h shape is [s, b, h, d]
    try:
        check_tensor_dim(q_h, 4)
        check_tensor_dim(k_h, 4)
    except (TypeError, ValueError) as e:
        logger.warning(f"Calculate qk tensor failed: {e}")
        return torch.tensor(0)

    if order == "s,b,h,d":
        qkt = torch.matmul(
            q_h[:, 0, 0, :], k_h[:, 0, 0, :].t()) / q_h.shape[-1] ** 0.5
    elif order == "b,s,h,d":
        qkt = torch.matmul(
            q_h[0, :, 0, :], k_h[0, :, 0, :].t()) / q_h.shape[-1] ** 0.5
    else:
        logger.warning("Calculate qk tensor failed: Order unsupported.")
        qkt = torch.tensor(0)
    return qkt


@torch.no_grad()
def cal_stable_rank(weight: torch.Tensor):
    eig = max_eigenvalue(weight)
    if eig == torch.tensor(0):
        return torch.tensor(0), torch.tensor(0)
    f_norm = torch.norm(weight, p="fro")
    return f_norm / eig, eig



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
AlgorithmParameter class.
"""


class AlgorithmParameter:
    """
    The class for algorithm parameter
    """

    def __init__(self: any) -> None:
        pass

    def get(self: any) -> dict:
        """
        get dict
        """
        return self.__dict__

    def get_dict(self: any) -> dict:
        """
        get dict
        """
        return self.__dict__



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from gzip import FEXTRA
import os
import re
import uuid
from collections import defaultdict
from datetime import datetime
from functools import partial

import pytz
import pandas as pd
import mindspore
from mindspore import Tensor, mint
from mindspore import nn, _no_grad

from msprobe.core.common.log import logger
from msprobe.core.common.const import MonitorConst, Const
from msprobe.core.common.file_utils import load_json, save_json, create_directory
from msprobe.core.monitor.utils import validate_config, get_output_base_dir, get_target_output_dir
from msprobe.core.monitor.anomaly_processor import AnomalyScanner, AnomalyDataFactory, AnomalyDataWriter
from msprobe.mindspore.common.utils import is_mindtorch
from msprobe.mindspore.monitor.common_func import is_valid_instance, get_parameters, get_submodules, get_rank, \
    comm_is_initialized
from msprobe.mindspore.monitor.utils import get_summary_writer_tag_name, step_accumulates_one, is_skip_step, \
    get_metrics, get_entropy_metric, get_sr_metric
from msprobe.mindspore.monitor.optimizer_collect import OptimizerMonFactory
from msprobe.mindspore.monitor.data_writers import CSVWriterWithAD, BaseWriterWithAD, WriterInput
from msprobe.mindspore.monitor.distributed.wrap_distributed import api_register, create_hooks, op_aggregate
from msprobe.mindspore.monitor.features import cal_qkt
from msprobe.core.common.file_utils import write_df_to_csv
from msprobe.core.common.utils import analyze_api_call_stack

FORMAT_MAPPING = {
    MonitorConst.CSV: CSVWriterWithAD,
    MonitorConst.API: BaseWriterWithAD
}


def get_output_base_dir():
    return os.getenv(MonitorConst.MONITOR_OUTPUT_DIR, MonitorConst.DEFAULT_MONITOR_OUTPUT_DIR)


def get_param_struct(param):
    res = {}
    if isinstance(param, (tuple, list)):
        res['config'] = f'{type(param).__name__}[{len(param)}]'
        for i, x in enumerate(param):
            res[i] = f'size={tuple(x.shape)}, dtype={x.dtype}' if isinstance(x, Tensor) else f'{type(x)}'
    elif isinstance(param, Tensor):
        res['config'] = 'tensor'
        res['tensor'] = f'size={tuple(param.shape)}, dtype={param.dtype}'
    else:
        res['config'] = f'{type(param)}'
        logger.warning(f'Not support type({type(param)}) now, please check the type of param {param}')
    return res


def param_is_not_tensor_parallel_duplicate(param, tp_group):
    return (hasattr(param, 'tensor_model_parallel') and param.tensor_model_parallel) or (
            mint.distributed.get_rank(group=tp_group) == 0
    )


def param_is_data_parallel_duplicate(dp_group):
    return mint.distributed.get_rank(group=dp_group) != 0


def squash_param_name(param_name):
    for pattern in ['^.*\.(layers?\..*)', '^.*\.(embeddings?\..*)', '^.*\.(final.*)', '^.*\.(output.*)',
                    '^.*\.(norm.*)']:
        match = re.findall(pattern, param_name)
        if match:
            return match[0]
    return param_name


def is_recording_module(module_name, l2_targets, vpp_stage):
    if len(l2_targets) > 0:
        for pattern in [vpp_stage + squash_param_name(module_name), vpp_stage + module_name]:
            if pattern in l2_targets:
                return pattern
        return ""
    else:
        raise NotImplementedError("If monitering l2_features, the targets should be set specifically.")


# Used For Module Forward & Backward Collect
class ModuleHookContext:
    def __init__(self, module_name) -> None:
        self.step = 0
        self.micro_step = 0
        self.actv = defaultdict(dict)
        self.actvgrad = []
        self.module_name = module_name
        self.struct = {}
        self.stack = ""

    def reset(self):
        self.actv.clear()
        self.actvgrad.clear()


class FeatureHookContext:
    def __init__(self, module_name):
        self.step = 0
        self.micro_step = 0
        self.attention_feature = {}
        self.linear_feature = {}
        self.module_name = module_name

    def reset(self):
        self.attention_feature.clear()
        self.linear_feature.clear()


start_step = 0


# Used For Optimizer Weight Grad & M/V Collect
class OptimizerContext:
    def __init__(self) -> None:
        self.step = start_step
        self.param_mg_direction = defaultdict(float)
        self.param_adam_update = defaultdict()
        self.param_adam_ratio = defaultdict()
        self.param_weight_grad = defaultdict()
        self.param_exp_avg = defaultdict()
        self.exp_avg_metric = {}
        self.param_exp_avg_sq = defaultdict()
        self.exp_avg_sq_metric = {}
        self.metric_dict = {}
        self.param_metric = {}

    def reset(self) -> None:
        self.param_mg_direction.clear()
        self.param_adam_update.clear()
        self.param_adam_ratio.clear()
        self.param_weight_grad.clear()
        self.param_exp_avg.clear()
        self.exp_avg_metric.clear()
        self.param_exp_avg_sq.clear()
        self.exp_avg_sq_metric.clear()
        self.metric_dict.clear()
        self.param_metric.clear()


# Used For Weight Grad Collect
class GradContext:
    def __init__(self) -> None:
        self.pre = {}
        self.post = {}
        self.actv = {}

    def reset(self):
        self.pre.clear()
        self.post.clear()
        self.actv.clear()


class CommunicationContext:
    def __init__(self) -> None:
        self.data = {}

    @staticmethod
    def _agg(data):
        aggregated_data = {}
        for tag, op2tensorlist in data.items():
            aggregated_data[tag] = {}
            for op, tensorlist in op2tensorlist.items():
                aggregated_data[tag][op] = op_aggregate(op, tensorlist)
        return aggregated_data

    def reset(self):
        self.data = {}

    def aggregate(self):
        self.data = self._agg(self.data)


class TrainerMon:
    def __init__(self, config_file_path, process_group=None, params_have_main_grad=True) -> None:
        # TYPE1: 只在这里初始化的变量, 不会随着训练中途config配置改变而重置
        self.config_file_path = config_file_path
        self.process_group = process_group
        self.params_have_main_grad = params_have_main_grad
        self.is_mindtorch = is_mindtorch()
        self.config_timestamp = 0  # 后面有校验时间戳, 首次监测无需为了更新config文件时间戳而去改, 可通过dynamic_on开关直接打开
        self.config = load_json(config_file_path)
        validate_config(self.config)

        local_tz = pytz.timezone("Asia/Shanghai")  # 根据需要调整为目标时区
        cur_time = datetime.now(local_tz).strftime('%b%d_%H-%M-%S')
        self.unique_id = str(uuid.uuid4())[:8]
        self.output_base_dir = get_output_base_dir()
        time_tags = self.config.get("append_output", [])
        try:
            self.rank = get_rank()
            if time_tags:
                output_append_dirs = get_target_output_dir(self.output_base_dir, time_tags[0], time_tags[1])
                if str(self.rank) in output_append_dirs:
                    self.tensorboard_dir = output_append_dirs[str(self.rank)]
                    logger.info(f"Append rank({self.rank}) result to {self.tensorboard_dir}")
            else:
                self.tensorboard_dir = os.path.join(self.output_base_dir,
                                                    f"{cur_time}-rank{self.rank}-{self.unique_id}")
        except Exception as e:
            self.rank = 0
            self.tensorboard_dir = os.path.join(self.output_base_dir, f"{cur_time}-rank{self.rank}-{self.unique_id}")

        self.pp_stage = 0
        self.group_mates = [0]

        # TYPE2: 只会在set_monitor()主调中赋值的变量
        self.model = None
        self.vpp = False
        self.dp_group = None
        self.tp_group = None
        self.micro_batch_number = 1
        self.optimizer_mon = None

        # TYPE3: 会随着训练中途config配置更新或监测状态改变而重置的变量
        self.module_fwd_hook_context_by_module = defaultdict(ModuleHookContext)
        self.module_bwd_hook_context_by_module = defaultdict(ModuleHookContext)
        self.feature_hook_context_by_module = defaultdict(FeatureHookContext)
        self.optimizer_context = defaultdict(OptimizerContext)
        self.cc_context = defaultdict(CommunicationContext)
        self.grad_context = GradContext()
        self.handles = defaultdict(list)
        self.param2name = defaultdict(str)
        self.name2index = defaultdict()
        self.name2indices = defaultdict()
        self.name2param = {}
        self.duplicate_param = {}
        self.name2tag = {}
        self.param_name_call_id = {}
        self.call_id = 0
        self.module_struct = defaultdict(dict)
        self.weight_hooked = False
        self.optimizer_hooked = False
        self.param_registered = False
        self.struct_printed = False
        self.pre_step_hooks = []
        self.post_step_hooks = []

        # 动静态区分
        self.dynamic_enable = os.getenv("DYNAMIC_MONITOR", 'False').lower() == 'true'
        if self.dynamic_enable:
            logger.warning(f"DYNAMIC_MONITOR is set, "
                           f"please make sure you have 'dynamic_on' and 'collect_times' in {self.config_file_path}")
            self.monitoring = False
        else:
            self.set_config()
            # 静态且collect_times>0时在第0步self.monitoring就可以True, 动态默认在下一步开启
            if self.collect_times > 0:
                self.monitoring = True

    @staticmethod
    def get_linear_hook_target(module):
        if isinstance(module, nn.Embedding):
            return ''
        if hasattr(module, "num_embeddings") or hasattr(module, "vocab_start_index"):
            return ''
        for weight_name in ["weight", "wg"]:
            if hasattr(module, weight_name) and isinstance(getattr(module, weight_name), Tensor):
                if getattr(module, weight_name).dim() == 2:
                    return weight_name
        return ''

    def set_config(self):
        self.start_step = self.config.get("start_step", 0)
        self.collect_times = self.config.get("collect_times", 100000000)  # 默认大值, 目的是一直采集
        self.step_interval = self.config.get("step_interval", 1)
        self.has_collect_times = 0  # 重设采集计数器
        self.print_struct = self.config.get("print_struct", False)
        self.targets = self.config.get("targets", None)
        self.module_rank_list = self.config.get("module_ranks", [])
        self.format = self.config.get('format', MonitorConst.CSV)  # only csv supported in mindspore
        self.eps = self.config.get('eps', 1e-8)
        self.ops = self.config.get('ops', [])  # monitor mean/max/norm/min/nan...
        self.ndigits = self.config.get('ndigits', 6)
        self.all_xy = self.config.get('all_xy', False)
        self.xy_distribution = self.config.get('xy_distribution', False)
        self.forward_only = self.config.get('forward_only', False)
        self.backward_only = self.config.get('backward_only', False)
        self.ur_distribution = self.config.get('ur_distribution', False)  # vector and ratio vector of adam
        self.mv_distribution = self.config.get("mv_distribution", False)  # m/v of adam
        self.wg_distribution = self.config.get("wg_distribution", False)
        self.param_distribution = self.config.get("param_distribution", False)
        self.mg_direction = self.config.get('mg_direction', False)  # main grad direction
        self.cc_distribution = self.config.get("cc_distribution", {})  # communication ops
        self.stack_info = self.config.get('stack_info', False)
        self.monitor_mbs_grad = self.config.get('monitor_mbs_grad', False)
        self.recording_l2_features = self.config.get('recording_l2_features', False)
        self.sa_order = self.config.get('sa_order', "s,b,h,d")

        if not self.cc_distribution.get('enable', False):
            self.cc_log_only = False
        else:
            self.cc_codeline = self.cc_distribution.get('cc_codeline', [])
            self.cc_log_only = self.cc_distribution.get('cc_log_only', False)
            self.cc_logged_stack = defaultdict(set)
            self.cc_pre_hook = self.cc_distribution.get('cc_pre_hook', False)
        self.common_info()

        # 初始化AnomalyData工厂
        alert_setting = self.config.get('alert', {"rules": []})
        self.alert_rules = AnomalyScanner.load_rules(alert_setting["rules"])
        self.anomaly_data_factory = None
        if alert_setting.get('dump', False):
            self.anomaly_data_factory = AnomalyDataFactory(self.rank, self.pp_stage, self.group_mates)

        # 初始化writer, 创建输出目录
        if self.format not in FORMAT_MAPPING:
            logger.error(f"Unsupported format: {self.format}, use default format: {MonitorConst.CSV}")
            self.format = MonitorConst.CSV
        self.step_count_per_record = self.config.get('step_count_per_record', 1)
        if not self.module_rank_list or (self.rank in self.module_rank_list):
            writer = FORMAT_MAPPING[self.format]
            self.summary_writer = writer(
                WriterInput(
                    self.tensorboard_dir,
                    self.alert_rules,
                    self.unique_id,
                    self.anomaly_data_factory,
                    self.ndigits,
                    self.step_count_per_record
                )
            )

            # 初始化anomaly detected文件目录
            if self.anomaly_data_factory:
                self.anomaly_data_writer = AnomalyDataWriter(os.path.join(self.output_base_dir, "anomaly_detected"),
                                                             self.rank)
                self.anomaly_data_writer.init_detected_json()

    def common_info(self):
        if not self.xy_distribution:
            logger.info("> module input/output input_grad/output_grad is not monitored. ")
        if self.forward_only:
            logger.info("> only module forward is monitored. ")
        if not self.ur_distribution:
            logger.info("> update vector and ratio vector of adam is not monitored. ")
        if not self.mv_distribution:
            logger.info("> momentum and variance of adam is not monitored. ")
        if not self.wg_distribution:
            logger.info("> weight grad of specified module is not monitored. ")
        if not self.recording_l2_features:
            logger.info("> l2 features of specified module is not monitored. ")
        if not self.mg_direction:
            logger.info('> grad and momentum direction will not be compared.')
        if not self.cc_distribution.get('enable', False):
            logger.info("> cc operator is not monitored.")

    def set_monitor(
            self,
            model,
            optimizer,
            grad_acc_steps=1,
            tp_group=None,
            dp_group=None,
            start_iteration=0
    ):
        global start_step
        start_step = start_iteration
        self.micro_batch_number = grad_acc_steps
        self.dp_group = dp_group
        self.tp_group = tp_group
        self.optimizer_mon = OptimizerMonFactory.create_optimizer_mon(optimizer)
        self.hook_step_final(optimizer)
        if not isinstance(model, list):
            model = [model]
        self.model = model
        if len(model) > 1:
            self.vpp = True
            logger.info('vpp enabled')
        if not self.dynamic_enable:
            self.register_hooks(optimizer)

    def hook_step_final(self, optimizer):
        def step_final_hook(optimizer, *args, **kwargs):
            context = self.optimizer_context[optimizer]
            # 静态在第0步就可以保存, 动态在第0步不可以, 因为动态设计的就是重置后下一步开启, 第0步的self.monitoring还是False
            if self.monitoring:
                module_rank_valid = self.is_target_rank()
                step_condition = (context.step >= self.start_step and (
                        context.step - self.start_step) % self.step_interval == 0)
                if module_rank_valid and step_condition:
                    self.has_collect_times += 1

                    if self.anomaly_data_factory:
                        self.anomaly_data_factory.set_call_id(self.param_name_call_id)
                    self.write_xy_tb(context.step)
                    self.write_grad_tb(context.step)
                    self.write_mv_tb(context)
                    self.write_param_tb(context)
                    self.write_features_tb(context.step)
                    if self.stack_info:
                        self.write_stack_info()
                        self.stack_info = False
                        for handle in self.handles["stack"]:
                            handle.remove()
                        self.handles["stack"].clear()

                    if context.metric_dict:
                        self.summary_writer.write_metrics(self.ops, context.metric_dict, context.step, 'other')
                    context.metric_dict.clear()

                    if self.anomaly_data_factory:
                        self.anomaly_data_writer.write_detected_json(self.summary_writer.get_anomalies())
                    self.summary_writer.clear_anomalies()

                    self.call_id = 0
                    self.param_name_call_id.clear()

                    if self.has_collect_times >= self.collect_times:
                        self._remove_all_hooks_final(optimizer)

            context.step += 1
            self.dynamic_monitor(optimizer)

        def patch_step(func, optimizer):
            def wrapper(*args, **kwargs):
                for hook in self.pre_step_hooks:
                    hook(optimizer, args, kwargs)
                out = func(*args, **kwargs)
                for hook in self.post_step_hooks:
                    hook(optimizer, args, kwargs)
                step_final_hook(optimizer, args, kwargs)
                return out

            return wrapper

        if self.is_mindtorch:
            optimizer.__class__.step = patch_step(optimizer.__class__.step, optimizer)
        else:
            optimizer.__class__.construct = patch_step(optimizer.__class__.construct, optimizer)

        return

    def dynamic_monitor(self, optimizer):
        """
        If dynamic monitor enabled and config.json updated,
        remove hooks and register new hooks according to new configuration.
        """
        context = self.optimizer_context[optimizer]
        if not self.dynamic_enable:
            return
        try:
            # 如果文件时间戳没变, 可以不读取节省时间
            config_timestamp = os.path.getmtime(self.config_file_path)
            if config_timestamp == self.config_timestamp:
                return
            # 更新config文件最新修改时间戳
            self.config_timestamp = config_timestamp
            config = load_json(self.config_file_path)
        except Exception as e:
            logger.error(f"get config.json wrong because {e}, not updated, please check!!!")
            return

        if config.get("dynamic_on", False):
            try:
                validate_config(config)
                self.config = config
                self.set_config()
                self.start_step = context.step  # 动态启停时不受原start_step影响，永远从下一步开始
                logger.warning(f"config is updated at step{context.step - 1}, "
                               f"will start new hook at step{context.step}.")
            except Exception as e:
                logger.error(f"set config wrong because {e}, not updated, please check!!!")
                return

            self._remove_all_hooks(optimizer)
            self.register_hooks(optimizer)

    def register_hooks(self, optimizer):
        self._register_param_name()
        self.hook_modules()
        self.hook_optimizer(optimizer)
        self._patch_grad_sync()
        if self.cc_distribution.get('enable', False):
            self.handles['cc'] = api_register.initialize_hook(*create_hooks(context=self.cc_context, monitor=self))
            api_register.redirect_api()
        self.monitoring = True

    def hook_modules(self):
        if not self.is_target_rank():
            return
        module_in_all_stage = [key for key in self.targets.keys() if MonitorConst.NAME_SEP not in key]

        for key in module_in_all_stage:
            struct = self.targets.pop(key)
            self.targets.update(
                {f'{vpp_stage}{MonitorConst.NAME_SEP}{key}': struct for vpp_stage in range(len(self.model))})

        hooked_count = 0
        for vpp_stage, model_chunk in enumerate(self.model):
            if not is_valid_instance(model_chunk):
                logger.info("Target Model is not Cell")
                continue
            vpp_stage = f'{vpp_stage}{MonitorConst.NAME_SEP}'
            targets = [x for x, _ in get_submodules(model_chunk)] if self.print_struct else self.targets.keys()
            l2_target_names = self.config.get('l2_targets', {})
            hooked_count += self._hook_module(targets, l2_target_names, model_chunk, vpp_stage)
        logger.info(f"> {hooked_count} modules are monitored.")

    def hook_optimizer(self, optimizer):
        def optimizer_pre_step_hook(opt, *args, **kwargs):
            context = self.optimizer_context[opt]
            if (self.print_struct and not all(value == {} for value in self.module_struct.values())
                    and not self.struct_printed):
                self._save_module_struct()
                if not self.cc_log_only:
                    raise Exception("exit after first monitor step when print model struct")
            if is_skip_step(context.step, self.start_step, self.step_interval, self.has_collect_times,
                            self.collect_times):
                return

            grad_dict = {}
            if self.wg_distribution:
                grad_dict = self.optimizer_mon.fetch_grad(self, self.param2name)

            if self.mv_distribution or self.ur_distribution or self.mg_direction:
                if self.is_mindtorch:
                    context.param_exp_avg, context.param_exp_avg_sq, context.param_adam_update, \
                        context.param_adam_ratio = self.optimizer_mon.fetch_mv(self, self.param2name)
                else:
                    context.param_exp_avg, context.param_exp_avg_sq = self.get_mv_for_ms(optimizer)

            self.generate_wgrad_metrics(grad_dict)
            self.generate_mv_metrics(context)
            self.generate_param_metrics(context, MonitorConst.PRE_PARAM)

            metric_dict = {}
            for cc in self.cc_context.values():
                cc.aggregate()
                metric_dict.update(cc.data)
                cc.reset()

            if not metric_dict:
                return
            context.metric_dict = metric_dict
            return

        def optimizer_post_step_hook(optimizer, args, kwargs):
            context = self.optimizer_context[optimizer]
            self.generate_param_metrics(context, MonitorConst.POST_PARAM)

        if self.optimizer_hooked or not self.is_target_rank():
            return

        self.pre_step_hooks.append(optimizer_pre_step_hook)
        self.post_step_hooks.append(optimizer_post_step_hook)
        self.optimizer_hooked = True
        return

    def generate_wgrad_metrics(self, grad_dict):
        if not self.wg_distribution:
            return

        get_metrics(self.ops, grad_dict, self.eps, self.grad_context.post)

    def generate_param_map(self, tag, param_tensor):
        metrics = {}
        if not self.is_mindtorch:
            return param_tensor
        for name in self.param2name.values():
            key = get_summary_writer_tag_name(name, tag, self.rank)
            self.register_param_call_id("optimizer_pre_step_hook", key)
            if name not in param_tensor or param_tensor[name] is None:
                continue
            metrics[key] = param_tensor[name]
        return metrics

    def generate_param_metrics(self, opt_context, stage=MonitorConst.PRE_PARAM):
        if not self.param_distribution:
            return
        tag2param = {
            self.name2tag.get(name, {}).get(stage): param
            for name, param in self.name2param.items()
            if param.numel() != 0
        }
        get_metrics(self.ops, tag2param, self.eps, opt_context.param_metric)

    def get_mv_for_ms(self, opt):
        if not self.mv_distribution:
            return {}, {}
        common_opt = opt
        if not is_valid_instance(opt):
            common_opt = getattr(opt, 'optimizer')
            if not is_valid_instance(common_opt):
                logger.warning("Optimizer is not valid, please check usage")
                return {}, {}
        m_dict = {}
        v_dict = {}
        for name, param in get_parameters(common_opt):
            if MonitorConst.EXP_AVG_SQ in name:
                v_dict[name] = param
            elif MonitorConst.EXP_AVG in name:
                m_dict[name] = param
        return m_dict, v_dict

    def generate_mv_metrics(self, opt_context):
        if not self.mv_distribution:
            return
        opt_context.exp_avg_metric = {}
        opt_context.exp_avg_sq_metric = {}
        m_tag_tensor_map = self.generate_param_map(MonitorConst.EXP_AVG, opt_context.param_exp_avg)
        v_tag_tensor_map = self.generate_param_map(MonitorConst.EXP_AVG_SQ, opt_context.param_exp_avg_sq)
        get_metrics(self.ops, m_tag_tensor_map, self.eps, opt_context.exp_avg_metric)
        get_metrics(self.ops, v_tag_tensor_map, self.eps, opt_context.exp_avg_sq_metric)

    def write_stack_info(self):
        stack_data = []
        header = ["module_name", "stack_info"]
        stack_data.append(header)
        for _, fwd_context in self.module_fwd_hook_context_by_module.items():
            stack_data.append([fwd_context.module_name, fwd_context.stack])
        filepath = os.path.join(self.tensorboard_dir, f'stack_info.csv')
        if not os.path.exists(filepath):
            data_frame = pd.DataFrame(columns=stack_data)
            write_df_to_csv(data_frame, filepath)

    def write_xy_tb(self, step):
        if not self.xy_distribution:
            return
        for _, fwd_context in self.module_fwd_hook_context_by_module.items():
            if len(fwd_context.actv) == 0:
                continue
            self.summary_writer.write_metrics(self.ops, fwd_context.actv, step, MonitorConst.ACTV)
            fwd_context.actv.clear()
        if self.grad_context.actv:
            self.summary_writer.write_metrics(self.ops, self.grad_context.actv, step, MonitorConst.ACTVGRAD)

    def write_param_tb(self, opt_context):
        if not self.param_distribution:
            return
        param_metrics = {k: v for k, v in opt_context.param_metric.items() if MonitorConst.PRE_PARAM in k}
        updated_param_metrics = {k: v for k, v in opt_context.param_metric.items() if MonitorConst.POST_PARAM in k}
        self.summary_writer.write_metrics(self.ops, param_metrics, opt_context.step, MonitorConst.PRE_PARAM)
        self.summary_writer.write_metrics(self.ops, updated_param_metrics, opt_context.step, MonitorConst.POST_PARAM)

    def write_mv_tb(self, opt_context):
        if not self.mv_distribution:
            return
        self.summary_writer.write_metrics(self.ops, opt_context.exp_avg_metric, opt_context.step, MonitorConst.EXP_AVG)
        self.summary_writer.write_metrics(self.ops, opt_context.exp_avg_sq_metric, opt_context.step,
                                          MonitorConst.EXP_AVG_SQ)

    def write_grad_tb(self, step):
        if not self.wg_distribution:
            return

        self.summary_writer.write_metrics(self.ops, self.grad_context.pre, step, 'grad_unreduced',
                                          use_micro_step=self.monitor_mbs_grad)
        self.summary_writer.write_metrics(self.ops, self.grad_context.post, step, 'grad_reduced')

    def write_metrics_if_not_empty(self, features, metrics, step, hook_name):
        if not features or len(features) == 0:
            return
        use_micro_step = hook_name not in ["linear_hook"]
        self.summary_writer.write_metrics(metrics, features, step, hook_name, use_micro_step=use_micro_step)
        features.clear()

    def write_features_tb(self, step):
        if not self.recording_l2_features:
            return
        for context in self.feature_hook_context_by_module.values():
            num_features = len(context.attention_feature) + len(context.linear_feature)
            if num_features == 0:
                continue
            self.write_metrics_if_not_empty(context.attention_feature, ["entropy", "softmax"], step,
                                            "attention_hook")
            self.write_metrics_if_not_empty(context.linear_feature, ["sr", "kernel_norm"], step,
                                            "linear_hook")

    def is_target_rank(self):
        if self.module_rank_list and (self.rank not in self.module_rank_list):
            return False
        return True

    def build_tbtag_tensor_map(self, module_name, suffix, tag, tensor):
        """
        :param module_name: str of module name
        :param suffix:
        :param tag:
        :param tensor: torch.tensor or tuple/list of torch.tensor
        :return: tensor_map
        """
        tensor_map = {}
        if isinstance(tensor, Tensor):
            tensor = [tensor]
        if isinstance(tensor, tuple) or isinstance(tensor, list):
            if len(tensor) == 1:
                key = get_summary_writer_tag_name(module_name + suffix, tag, self.rank)
                self.register_param_call_id("_hook_module", key)
                tensor_map[key] = tensor[0]
            else:
                for i, tensor_i in enumerate(tensor):
                    key = get_summary_writer_tag_name(module_name + f"_{i}" + suffix, tag, self.rank)
                    self.register_param_call_id("_hook_module", key)
                    tensor_map[key] = tensor_i
        return tensor_map

    def register_param_call_id(self, hook_name: str, key: str):
        """
        :param hook_name:
        :param key: str, '0:relu_0/output_grad'
        :return:
        """
        logger.debug(f"{hook_name} {key}: {self.call_id}")
        self.param_name_call_id[key] = self.call_id
        self.call_id += 1

    def _register_param_name(self):
        for vpp_stage, model_chunk in enumerate(self.model):
            prefix = f'{vpp_stage}{MonitorConst.NAME_SEP}'
            self._register_chunk(model_chunk, prefix)

    def _register_chunk(self, model_chunk, prefix):
        index = 0
        for param_name, param in get_parameters(model_chunk):
            if not param.requires_grad:
                continue
            if self._is_target_param(param_name, param, prefix):
                name = prefix + squash_param_name(param_name)
                if name in self.param2name.values():
                    name = prefix + param_name
                self.param2name[param] = name
                self.name2param[name] = param
                self.name2index[name] = index

                if self.tp_group and not param_is_not_tensor_parallel_duplicate(param, self.tp_group):
                    self.duplicate_param[name] = True
                if self.dp_group and param_is_data_parallel_duplicate(self.dp_group):
                    self.duplicate_param[name] = True
                keywords = [
                    MonitorConst.PRE_GRAD,
                    MonitorConst.POST_GRAD,
                    MonitorConst.PRE_PARAM,
                    MonitorConst.POST_PARAM
                ]
                self.name2tag[name] = {
                    k: get_summary_writer_tag_name(name, k, self.rank)
                    for k in keywords
                }
                index += 1

    def _save_module_struct(self):
        output_dir = os.path.join(get_output_base_dir(), 'module_struct', f'rank{self.rank}')
        create_directory(output_dir)
        module_struct_file = os.path.realpath(os.path.join(output_dir, 'module_struct.json'))
        save_json(module_struct_file, self.module_struct, indent=2)
        logger.info(f"> save module struct to {module_struct_file}")
        self.struct_printed = True

    def _hook_module(self, target_names, l2_target_names, module, vpp_stage=''):
        if not is_valid_instance(module):
            # nothing to hook
            return 0

        def fwd_hook_fun(module, args, kwargs, module_output, name):

            module_input = [tensor for tensor in args if isinstance(tensor, Tensor)]
            if kwargs:
                kwargs_tensors = [tensor for tensor in kwargs.values() if isinstance(tensor, Tensor)]
                module_input.extend(kwargs_tensors)

            if module not in self.module_fwd_hook_context_by_module:
                self.module_fwd_hook_context_by_module[module] = ModuleHookContext(name)
            context: ModuleHookContext = self.module_fwd_hook_context_by_module[module]
            if not context.struct:
                context.struct = {
                    Const.INPUT: get_param_struct(module_input),
                    Const.OUTPUT: get_param_struct(module_output)
                }
            if self.print_struct:
                self.module_struct[context.module_name].update(context.struct)
                return
            if not module.training:
                return
            if is_skip_step(context.step, self.start_step, self.step_interval, self.has_collect_times,
                            self.collect_times):
                step_accumulates_one(context, self.micro_batch_number)
                return

            tbtag_tensor_map = {}
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.INPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTV, module_input))
            module_output = [tensor for tensor in module_output if isinstance(tensor, Tensor)] \
                if isinstance(module_output, tuple) else module_output
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.OUTPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTV, module_output))
            try:
                get_metrics(self.ops, tbtag_tensor_map, self.eps, context.actv)
            except Exception as e:
                logger.warning(f"An error occurred while generating forward activation metrics")

            step_accumulates_one(context, self.micro_batch_number)
            return

        def bwd_hook_fun(module, input_grad, output_grad):
            context: ModuleHookContext = self.module_bwd_hook_context_by_module[module]
            if not context.struct:
                context.struct = {
                    MonitorConst.ACTVGRAD_IN: get_param_struct(input_grad),
                    MonitorConst.ACTVGRAD_OUT: get_param_struct(output_grad)
                }
            if self.print_struct:
                self.module_struct[context.module_name].update(context.struct)
                return

            if is_skip_step(context.step, self.start_step, self.step_interval, self.has_collect_times,
                            self.collect_times):
                step_accumulates_one(context, self.micro_batch_number)
                return

            valid_input_grad = [tensor for tensor in input_grad if isinstance(tensor, Tensor)]
            tbtag_tensor_map = {}
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.INPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTVGRAD, valid_input_grad))

            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.{Const.OUTPUT}', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    MonitorConst.ACTVGRAD, output_grad))

            if context.micro_step == 0 and context.actvgrad:
                logger.warning(f"actvgrad context of {context.module_name} is not empty when first micro_step, "
                               f"maybe something wrong happened. Now clear it.")
                context.actvgrad.clear()
            try:
                get_metrics(self.ops, tbtag_tensor_map, self.eps, self.grad_context.actv)
            except Exception as e:
                logger.warning(f"An error occurred while generating backward activation metrics: {e}")

            step_accumulates_one(context, self.micro_batch_number)
            return

        def fwd_hook_register(module, fwd_hook_fun, name):
            from packaging import version
            if version.parse(mindspore.__version__) >= version.parse('2.6.0'):
                def wrapper(module, args, kwargs, module_output):
                    return fwd_hook_fun(module, args, kwargs, module_output, name)

                return module.register_forward_hook(wrapper, with_kwargs=True)

            else:
                def wrapper(module, args, module_output):
                    return fwd_hook_fun(module, args, None, module_output, name)

                return module.register_forward_hook(wrapper)

        def extract_attention_feature_hook(module, args, kwargs, module_output, name):
            module_input = [tensor for tensor in args if isinstance(tensor, Tensor)]
            if kwargs:
                kwargs_tensors = [tensor for tensor in kwargs.values() if isinstance(tensor, Tensor)]
                module_input.extend(kwargs_tensors)

            if module not in self.feature_hook_context_by_module:
                self.feature_hook_context_by_module[module] = FeatureHookContext(name)
            context: FeatureHookContext = self.feature_hook_context_by_module[module]

            tbtag_tensor_map = {}
            if len(module_input) < 2:
                logger.warning(
                    "Calculate attention feature failed, the length of module_input in attention hook's module should "
                    "be greater than or equal to 2.")

            q_h = module_input[0]
            k_h = module_input[1]
            qkt = cal_qkt(q_h, k_h, order=self.sa_order)
            tbtag_tensor_map.update(
                self.build_tbtag_tensor_map(
                    f'{context.module_name}.attention', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                    'qkt', qkt))
            get_entropy_metric(tbtag_tensor_map, context.attention_feature)

            context.micro_step += 1
            if context.micro_step == self.micro_batch_number:
                context.micro_step = 0
                context.step += 1
            return

        def extract_linear_sr_hook(module, args, kwargs, module_output, name):
            weight_name = self.get_linear_hook_target(module)
            if weight_name == "":
                return
            if module not in self.feature_hook_context_by_module:
                self.feature_hook_context_by_module[module] = FeatureHookContext(name)
            context: FeatureHookContext = self.feature_hook_context_by_module[module]

            if context.micro_step == self.micro_batch_number - 1:
                tbtag_tensor_map = {}
                value = module.weight.data
                tbtag_tensor_map.update(
                    self.build_tbtag_tensor_map(
                        f'{context.module_name}.linear', f'{MonitorConst.NAME_SEP}{context.micro_step}',
                        'sr', value))

                get_sr_metric(tbtag_tensor_map, context.linear_feature)

            context.micro_step += 1
            if context.micro_step == self.micro_batch_number:
                context.micro_step = 0
                context.step += 1
            return

        def stack_hook(module, args, kwargs, module_output, name):
            if module not in self.module_fwd_hook_context_by_module:
                self.module_fwd_hook_context_by_module[module] = ModuleHookContext(name)
            context: ModuleHookContext = self.module_fwd_hook_context_by_module[module]
            context.stack = analyze_api_call_stack(name)
            return

        if self.backward_only and self.forward_only:
            logger.warning('not enable backward_only and forward_only simultaneously')
        hooked_count = 0

        for module_name, submodule in get_submodules(module):
            if self.stack_info:
                name = vpp_stage + squash_param_name(module_name)
                handle = fwd_hook_register(submodule, stack_hook, name=name)
                self.handles["stack"].append(handle)
            name = self._is_target_module(module_name, target_names, vpp_stage)
            if not name:
                continue
            if self.xy_distribution or self.print_struct:
                if not self.backward_only:
                    handle = fwd_hook_register(submodule, fwd_hook_fun, name=name)
                    self.handles['xy'].append(handle)
                if not self.forward_only:
                    handle = submodule.register_backward_hook(bwd_hook_fun)
                    self.handles['xy'].append(handle)
                    self.module_bwd_hook_context_by_module[submodule] = ModuleHookContext(name)
                logger.info(f"> {name} is monitored successfully")
                hooked_count += 1

        if not self.print_struct and self.recording_l2_features:
            for module_name, submodule in get_submodules(module):
                func_map = {
                    "attention_hook": extract_attention_feature_hook,
                    "linear_hook": extract_linear_sr_hook
                }
                for hook in func_map.keys():
                    if hook in l2_target_names:
                        temp_names = l2_target_names[hook]
                        name = is_recording_module(module_name, temp_names, vpp_stage)
                        if name:
                            handle = fwd_hook_register(submodule, func_map[hook], name=name)
                            print_feature_name = hook.split('_')[0]
                            logger.info_on_rank_0(
                                f'> {print_feature_name} features of {name} is monitored successfully')
                            self.handles["L2_features"].append(handle)
                            hooked_count += 1
        return hooked_count

    def _patch_grad_sync(self):
        if not self.wg_distribution:
            return
        self._hook_weights()

    def _hook_weights(self):
        context = self.grad_context

        @_no_grad()
        def param_hook(grad, param, name):
            key = name
            if self.monitor_mbs_grad:
                key += f'{MonitorConst.NAME_SEP}{param.micro_step}'
            key = get_summary_writer_tag_name(key, 'acc_grad', self.rank)
            self.register_param_call_id("param_hook", key)
            param.micro_step += 1
            if self.monitor_mbs_grad or (param.micro_step == self.micro_batch_number):
                get_metrics(self.ops, {key: grad}, self.eps, self.grad_context.pre)
            if param.micro_step == self.micro_batch_number:
                param.micro_step = 0

        def param_hook_wrapper(param_hook, param, name):
            def wrapper(grad):
                return param_hook(grad, param, name)

            return wrapper

        logger.info("hooking weight grads.")
        for param, name in self.param2name.items():
            setattr(param, 'micro_step', 0)
            handle = param.register_hook(
                param_hook_wrapper(param_hook, param=param, name=name))
            self.handles['wgrads'].append(handle)
        self.weight_hooked = True

    def _is_target_param(self, param_name, param, prefix):
        if not self.targets:
            return True
        squash_name = prefix + squash_param_name(param_name)
        name = prefix + param_name
        for target in self.targets.keys():
            if param_name.startswith(target) or squash_name.startswith(target) or name.startswith(target):
                setattr(param, "zero_out_wgrad", True)
                return True
        return False

    def _is_target_module(self, module_name, targets, vpp_stage):
        if self.all_xy or self.print_struct:
            return vpp_stage + squash_param_name(module_name)
        for pattern in [
            vpp_stage + squash_param_name(module_name),
            vpp_stage + module_name,
        ]:
            if pattern in targets:
                return pattern
        return ""

    def _remove_all_hooks(self, optimizer):
        # 清空hook handle
        for handle in self.handles['xy']:
            handle.remove()
        self.handles['xy'].clear()
        for handle in self.handles['L2_features']:
            handle.remove()
        self.handles['L2_features'].clear()
        # 清空对应context缓存
        for fwd_context in self.module_fwd_hook_context_by_module.values():
            fwd_context.reset()
        for bwd_context in self.module_bwd_hook_context_by_module.values():
            bwd_context.reset()
        for feature_context in self.feature_hook_context_by_module.values():
            feature_context.reset()
        self.grad_context.reset()  # 权重梯度和激活值梯度都在这

        for handle in self.handles['wgrads']:
            handle.remove()
        self.handles['wgrads'].clear()
        self.weight_hooked = False

        if self.optimizer_hooked:
            self.pre_step_hooks.clear()
            self.post_step_hooks.clear()
        for _, context in self.optimizer_context.items():
            context.reset()
        self.optimizer_hooked = False

        for handle in self.handles['cc']:
            handle.remove()
        self.handles['cc'].clear()
        api_register.restore_api()
        for _, context in self.cc_context.items():
            context.reset()

        # 清空节点缓存
        self.param2name.clear()
        self.name2index.clear()
        self.name2indices.clear()
        self.name2param.clear()
        self.duplicate_param.clear()
        self.name2tag.clear()
        self.module_struct.clear()

        # 关闭采集状态
        self.monitoring = False

    def _remove_all_hooks_final(self, optimizer):
        if self.dynamic_enable:
            # 结束后自动重置dynamic_on为False等待用户手动开启
            try:
                config = load_json(self.config_file_path)
                config['dynamic_on'] = False
                save_json(self.config_file_path, config, indent=2)
                config_timestamp = os.path.getmtime(self.config_file_path)
                self.config_timestamp = config_timestamp
                logger.info(
                    "Finish monitor, set config'dynamic_on=False, will restart by set it to True and update config")
            except Exception as e:
                logger.warning(f"Finish monitor, set config'dynamic_on=False fail because {e}, please check!!!")
        logger.info("Finish monitor")
        self._remove_all_hooks(optimizer)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from dataclasses import dataclass
from msprobe.core.common.const import Const
from msprobe.core.common.log import logger
from msprobe.core.common.const import CompareConst
from msprobe.core.compare.find_first.utils import RankPath, DiffAnalyseConst


@dataclass
class DataNode:
    op_name: str
    rank: int
    inputs: dict
    outputs: dict
    op_data: list
    layer: int = 0  # 和communication_node的layer保持一致
    sub_layer: int = 0  # 调用顺序，越小表示越先调用

    def __init__(self, op_name, rank, op_data, **kwargs):
        self.op_name = op_name
        self.rank = rank
        self.stack = None
        self.inputs = {}
        self.outputs = {}
        self.is_diff = False
        self.parse_data(op_data)
        self.sub_layer = kwargs.get('sub_layer', 0)

    def find_stack(self):
        for item in self.stack:
            if len(item) >= 2 and self.op_name in item[0]:
                return item[1]
        return {}

    def parse_data(self, op_data):
        self.is_diff = not op_data.get("is_same", True)
        self.op_data = op_data.get("op_items") # 这里拿到的是比对column，是一个list，有若干行
        metrics = {}
        for cmp_data in self.op_data:
            name = cmp_data.get(CompareConst.NPU_NAME)
            # 构建度量指标字典
            metrics = {}
            
            if CompareConst.NPU_MAX in cmp_data:
                metrics = {CompareConst.NPU_MAX: cmp_data.get(CompareConst.NPU_MAX),
                        CompareConst.NPU_MIN: cmp_data.get(CompareConst.NPU_MIN),
                        CompareConst.NPU_MEAN: cmp_data.get(CompareConst.NPU_MEAN),
                        CompareConst.NPU_NORM: cmp_data.get(CompareConst.NPU_NORM)}
            elif CompareConst.NPU_MD5 in cmp_data:
                metrics[CompareConst.NPU_MD5] = cmp_data.get(CompareConst.NPU_MD5)
                
            if CompareConst.NPU_P2POP_PEER in cmp_data:
                metrics[CompareConst.NPU_P2POP_PEER] = cmp_data.get(CompareConst.NPU_P2POP_PEER)

            if cmp_data.get(CompareConst.STACK) != CompareConst.N_A and not self.stack:
                self.stack = cmp_data.get(CompareConst.STACK)
            if cmp_data.get('state') == "input":
                self.inputs[name] = metrics
            elif cmp_data.get('state') == "output":
                self.outputs[name] = metrics

    def gen_node_info(self, path: RankPath):
        data_info_list = {Const.INPUT: self.inputs, Const.OUTPUT: self.outputs}
        return {'op_name': self.op_name,
                'data_info': data_info_list,
                'stack_info': self.stack}


class CommunicationNode:
    def __init__(self, node_id, rank, data: DataNode, layer=0, **kwargs):
        self.node_id = node_id
        self.rank = rank
        self.data = data
        self.is_diff = data.is_diff
        self.layer = layer
        op_name_split = self.data.op_name.split(Const.SEP)
        if len(op_name_split) < 4:
            logger.error(f'invalid op_name: {self.data.op_name}')
            raise RuntimeError(f'invalid op_name: {self.data.op_name}')
        self.api = op_name_split[1]
        self.call_cnt = op_name_split[2]
        self.pre_node = kwargs.get('pre_node')
        self.link_nodes = kwargs.get('link_nodes', {})
        self.dst_nodes = kwargs.get('dst_nodes', {})
        self.src_nodes = kwargs.get('src_nodes', {})
        self.next_nodes = kwargs.get('next_nodes', {})
        self.compute_ops = kwargs.get('compute_ops', [])
        self.type = self._resolve_type()
        self.connected = False

    def add_next(self, node):
        self.next_nodes[node.node_id] = node
        node.pre_node = self
        node.layer = self.layer + 1
        node.data.layer = node.layer

    def add_link(self, node):
        self.link_nodes[node.node_id] = node
        node.link_nodes[self.node_id] = self
        node.layer = self.layer
        node.data.layer = node.layer
        self.connected = True
        node.connected = True

    def add_dst(self, node):
        self.dst_nodes[node.node_id] = node
        node.src_nodes[self.node_id] = self
        node.layer = self.layer
        node.data.layer = node.layer
        self.connected = True
        node.connected = True

    def delete(self):
        for node in self.next_nodes.values():
            node.pre_node = None
        for node in self.dst_nodes.values():
            if node.src_nodes:
                node.src_nodes.pop(self.node_id)
        for node in self.src_nodes.values():
            if node.dst_nodes:
                node.dst_nodes.pop(self.node_id)
        for node in self.link_nodes.values():
            if node.link_nodes:
                node.link_nodes.pop(self.node_id)
        if self.pre_node:
            if self.pre_node.next_nodes:
                self.pre_node.next_nodes.pop(self.node_id)

    def find_connected_nodes(self):
        """
        根据 api/类型/入参/调用次数 确定相连接的node的op_name
        """
        tar_api = DiffAnalyseConst.P2P_API_MAPPING.get(self.api, self.api)
        ranks = set()
        # 遍历DST和SRC相关的input，获取对应的rank值
        # 遍历inputs获取所有rank值
        for input_name, v in self.data.inputs.items():
            # 检查key是否包含DST/SRC相关标识
            target_types = [DiffAnalyseConst.DST, DiffAnalyseConst.DST_GROUP,
                          DiffAnalyseConst.SRC, DiffAnalyseConst.SRC_GROUP]
            if any(keyword in input_name for keyword in target_types):
                # 优先使用MD5值，如果没有则使用NPU_MAX值
                rank_val = 0
                if CompareConst.NPU_MD5 in v:
                    rank_val = int(v.get(CompareConst.NPU_MD5, 0))
                else:
                    rank_val = int(v.get(CompareConst.NPU_MAX, 0))
                if rank_val:
                    ranks.add(rank_val)
            elif input_name.endswith('.group'):
                # 优先使用MD5值，如果没有则使用NPU_MAX值
                val = v.get(CompareConst.NPU_MD5) if CompareConst.NPU_MD5 in v else v.get(CompareConst.NPU_MAX)
                if val and val.startswith('[') and val.endswith(']'):
                    val = [int(part) for part in val.strip('[]').split(',')]
                    ranks.update(val)
            elif v.get(CompareConst.NPU_P2POP_PEER) != "None":
                ranks.add(v.get(CompareConst.NPU_P2POP_PEER))

        return {'ranks': ranks, 'api': f'Distributed.{tar_api}',
                'type': DiffAnalyseConst.OPPOSITE_DIR.get(self.type, DiffAnalyseConst.LINK)}


    def _resolve_type(self):
        # 遍历SRC和DST相关的输入，根据rank值判断节点类型
        for prefix, node_type in [(DiffAnalyseConst.SRC, DiffAnalyseConst.SRC), 
                                (DiffAnalyseConst.DST, DiffAnalyseConst.DST)]:
            for k, v in self.data.inputs.items():
                if prefix in k or f"{prefix}_GROUP" in k:
                    # 优先使用MD5值，如果没有则使用NPU_MAX值
                    compare_val = v.get(CompareConst.NPU_MD5) if CompareConst.NPU_MD5 in v \
                                  else v.get(CompareConst.NPU_MAX)
                    return node_type if compare_val == self.rank \
                           else DiffAnalyseConst.OPPOSITE_DIR[node_type]
        return DiffAnalyseConst.LINK



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import math
import os
import json
import re
import stat
import threading
from functools import cmp_to_key
from pathlib import Path
from tensorboard.util import tb_logging
from .global_state import GraphState
from .constant import DataType, FILE_NAME_REGEX, MAX_FILE_SIZE, PERM_GROUP_WRITE, PERM_OTHER_WRITE, COLOR_PATTERN
from .i18n import language, ZH_CN
# 创建一个全局锁
_thread_local_lock = threading.Lock()
logger = tb_logging.get_logger()
FILE_PATH_MAX_LENGTH = 4096


class GraphUtils:

    @staticmethod
    def t(key):
        lang = GraphState.get_global_value('lang', ZH_CN)
        return language.get(lang).get(key, '')

    @staticmethod
    def get_graph_data(meta_data):
        if not meta_data:
            return None, 'Error: no query parameters provided'
        try:
            run_name = meta_data.get('run')
            runs = GraphState.get_global_value('runs', {})
            run = runs.get(run_name) or run_name
            tag = meta_data.get('tag')
            current_tag = GraphState.get_global_value('current_tag')
            current_run = GraphState.get_global_value('current_run')
            if current_tag == tag and current_run == run:
                return GraphState.get_global_value('current_file_data'), None  # 直接返回获取结果
            else:
                json_data, error_message = GraphUtils.safe_load_data(run, f"{tag}.vis", False)
                if error_message:
                    return None, 'Error: fail to get graph data'
                GraphState.set_global_value('current_file_data', json_data)
                GraphState.set_global_value('current_tag', tag)
                GraphState.set_global_value('current_run', run)
                return json_data, error_message
        except Exception as e:
            logger.error(f'Error: fail to get graph data by {meta_data}, error: {e}')
            return None, 'Error: fail to get graph data'

    @staticmethod
    def get_opposite_node_name(node_name):
        opposite_node_name = ''
        # 如果npu_node_name包含forward，则opposite_npu_node_name为npu_node_name替换forward为backward
        if 'forward' in node_name:
            opposite_node_name = node_name.replace('forward', 'backward')
        else:
            opposite_node_name = node_name.replace('backward', 'forward')
        return opposite_node_name

    @staticmethod
    def get_parent_node_list(graph_data, node_name):
        """获取父节点列表"""
        # 如果 graph_data 为空或 node_name 不存在，直接返回空列表
        if not graph_data or not node_name:
            return []

        node_list = []  # 存储结果的列表
        current_node = node_name  # 当前节点名称
        nodes = graph_data['node']  # 提取节点数据，避免重复访问

        while current_node:
            node_list.append(current_node)  # 将当前节点加入列表
            # 获取当前节点的 upnode
            current_node = nodes.get(current_node, {}).get('upnode')

            # 检测循环引用（防止死循环）
            if current_node in node_list:
                raise ValueError(
                    f"{GraphUtils.t('circularReferenceError1')}"
                    f"{current_node}{GraphUtils.t('circularReferenceError2')}"
                )

        return list(reversed(node_list))  # 返回结果列表

    @staticmethod
    def split_graph_data_by_microstep(graph_data, micro_step):
        graph_nodes = graph_data.get('node', {})
        if str(micro_step) == str(-1):
            return graph_nodes
        splited_graph_data = {}
        node_list = []  # 存储已遍历的列表

        def traverse_npu(graph_nodes, subnodes):
            for node_name in subnodes:
                node_data = graph_nodes.get(node_name, {})
                micro_step_id = node_data.get('micro_step_id')
                if str(micro_step_id) == str(micro_step) or micro_step_id is None:
                    splited_graph_data[node_name] = (node_data)
                    if node_name in node_list:
                        raise ValueError(
                            f"{GraphUtils.t('circularReferenceError1')}"
                            f"{node_name}{GraphUtils.t('circularReferenceError2')}"
                        )
                    node_list.append(node_name)
                    traverse_npu(graph_nodes, node_data.get('subnodes', []))

        root = graph_data.get('root')
        root_subnodes = graph_nodes.get(root, {}).get('subnodes', [])
        node_list.append(root)
        traverse_npu(graph_nodes, root_subnodes)
        return splited_graph_data

    @staticmethod
    def walk_with_max_depth(logdir, max_depth):
        for root, dirs, files in os.walk(logdir):
            # 计算当前 root 相对于 top 的深度
            depth = root[len(logdir):].count(os.sep) + 1
            if depth >= max_depth:
                del dirs[:]
            yield root, dirs, files

    @staticmethod   
    def safe_json_loads(json_str, default_value=None):
        """
        安全地解析 JSON 字符串，带长度限制和异常处理。
        :param json_str: 要解析的 JSON 字符串
        :param default_value: 如果解析失败返回的默认值
        :return: 解析后的 Python 对象 或 default_value
        """
        # 类型检查
        if not isinstance(json_str, str):
            return default_value
        # 长度限制
        if len(json_str) > MAX_FILE_SIZE:
            return default_value
        try:
            result = json.loads(json_str)
            return result
        except json.JSONDecodeError as e:
            logger.error(f"JSON decode error: {e}")
            return default_value
        except Exception as e:
            logger.error(f"Unexpected error: {e}")
            return default_value

    @staticmethod
    def safe_get_node_info(data, default_value=None):
        node_info = data.get('nodeInfo')
        try:
            # 长度限制 - 检查字典转为字符串后的长度
            node_info_str = str(node_info)
            if len(node_info_str) > MAX_FILE_SIZE:
                logger.error(f"Input length exceeds {MAX_FILE_SIZE} characters.")
                return default_value
            # 验证必要字段是否存在
            required_fields = ["nodeName", "nodeType"]
            for field in required_fields:
                if field not in node_info:
                    logger.error(f"Field {field} is missing in metadata.")
                    return default_value
            return node_info
        except json.JSONDecodeError:
            logger.error("NodeInfo parameter is not in valid JSON format.")
            return default_value
        except Exception as e:
            logger.error(f"An error occurred while parsing the nodeInfo parameter: {str(e)}")
            return default_value

    @staticmethod
    def safe_get_meta_data(data, default_value=None):
        meta_data = data.get('metaData')
        try:
            # 长度限制
            meta_data_str = str(meta_data)
            if len(meta_data_str) > MAX_FILE_SIZE:
                logger.error(f"Input length exceeds {MAX_FILE_SIZE} characters.")
                return default_value
            # 验证必要字段是否存在
            required_fields = ["tag", "microStep", "run", "type", 'lang']
            for field in required_fields:
                if field not in meta_data:
                    logger.error(f"Field {field} is missing in metadata.")
                    return default_value
            config_info = GraphState.get_global_value('config_info', {})
            micro_step = meta_data.get('microStep')
            # 验证非必要字段如果存在,进行范围限制
            if config_info.get('microSteps'):
                if int(micro_step) != -1 and int(micro_step) not in list(range(config_info.get('microSteps'))):
                    logger.error(f"Field microStep {micro_step} is not in config_info.")
                    return default_value
            if meta_data.get('type') == DataType.DB.value and config_info:
                rank = meta_data.get('rank')
                step = meta_data.get('step')
                if config_info.get('ranks') and rank:
                    if rank not in config_info.get('ranks'):
                        logger.error(f"Field rank {rank} is not in config_info.")
                        return default_value
                if config_info.get('steps') and step:
                    if step not in config_info.get('steps'):
                        logger.error(f"Field step {step} is not in config_info.")
                        return default_value
            return meta_data
            
        except json.JSONDecodeError:
            logger.error("MetaData parameter is not in valid JSON format.")
            return default_value
        except Exception as e:
            logger.error(f"An error occurred while parsing the metatdata parameter: {str(e)}")
            return default_value

    @staticmethod
    def remove_prefix(node_data, prefix):
        if node_data is None:
            return {}
        return {k[len(prefix):] if k.startswith(prefix) else k: v for k, v in node_data.items()}

    @staticmethod
    def convert_to_float(value):
        try:
            if isinstance(value, str):
                # 处理'0.0%, 由于Mean小于1e-06, 建议不参考此相对误差，请参考绝对误差'和'0.0%'的情况
                value = value.split(',')[0]
                if value.endswith('%'):
                    value = value.replace('%', '').strip()
                    return float(value) / 100.0
            return float(value)
        except ValueError:
            return float('nan')

    @staticmethod
    def format_relative_err(value):
        """格式化相对误差为百分比，保留四位小数"""
        if value is None or math.isnan(value):
            return "NaN"
        else:
            return "{:.4%}".format(value)

    @staticmethod
    def nan_to_str(value):
        """将 NaN 转换为 'NaN' 字符串"""
        return "NaN" if math.isnan(value) else value

    @staticmethod
    def is_relative_to(path, base):
        abs_path = os.path.abspath(path)
        abs_base = os.path.abspath(base)
        return os.path.commonpath([abs_path, abs_base]) == str(abs_base)

    @staticmethod
    def bytes_to_human_readable(size_bytes, decimal_places=2):
        """
        将字节大小转换为更易读的格式（如 KB、MB、GB 等）。
        
        :param size_bytes: int 或 float，表示字节大小
        :param decimal_places: 保留的小数位数，默认为 2
        :return: str，人类可读的大小表示
        """
        if size_bytes == 0:
            return "0 B"

        units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
        unit_index = 0

        while size_bytes >= 1024 and unit_index < len(units) - 1:
            size_bytes /= 1024.0
            unit_index += 1

        return f"{size_bytes:.{decimal_places}f} {units[unit_index]}"

    @staticmethod
    def safe_save_data(data, run_name, tag):
        runs = GraphState.get_global_value('runs', {})
        run = runs.get(run_name) or run_name
        if not run or not tag:
            error_message = 'The query parameters "run" and "tag" are required'
            return None, error_message
        try:
            # 检查 tag 是否为合法文件名
            if not re.match(FILE_NAME_REGEX, tag):
                raise ValueError(f"{GraphUtils.t('invalidTag')}{tag}.")
            # 构建文件路径并标准化
            file_path = os.path.join(run, tag)
            # 目录安全校验
            success, error = GraphUtils.safe_check_save_file_path(run, True)
            if not success:
                raise PermissionError(error)
            # 文件安全校验
            success, error = GraphUtils.safe_check_save_file_path(file_path)
            if not success:
                raise PermissionError(error)
            # 尝试写入文件
            with _thread_local_lock:
                with open(file_path, "w", encoding="utf-8") as f:
                    json.dump(data, f, ensure_ascii=False, indent=4)
                    f.flush()  # 强制将缓冲区内容写入操作系统
                    os.fsync(f.fileno())  # 强制将缓冲区内容写入磁盘
                os.chmod(file_path, 0o640)
            # 最终校验（防御TOCTOU攻击）
            if os.path.islink(file_path):
                raise RuntimeError("The file has been replaced with a symbolic link")
            return True, None
        except (TypeError, ValueError) as e:
            logger.error(f"Invalid data: {e}")
            return None, 'Invalid data'
        except OSError as e:
            logger.error(f"Failed to create directory: {run}. Error: {e}\n")
            return None, 'failed to create directory '
        except Exception as e:
            logger.error(f'Error: File "{file_path}" is not accessible. Error: {e}')
            return None, 'failed to save file'

    @staticmethod
    def safe_load_data(run_name, tag, only_check=False):
        runs = GraphState.get_global_value('runs', {})
        run_dir = runs.get(str(run_name)) or run_name
        safe_base_dir = GraphState.get_global_value('logdir')
        """Load a single .vis file from a given directory based on the tag."""
        if run_dir is None or tag is None:
            error_message = 'The query parameters "run" and "tag" are required'
            return None, error_message
        try:
            file_path = os.path.join(run_dir, tag)
            # 安全验证：基础路径校验
            if not GraphUtils.is_relative_to(file_path, safe_base_dir):
                raise ValueError(GraphUtils.t('pathMayNotInSecureDirectory'))
            # 目录安全校验
            success, error = GraphUtils.safe_check_load_file_path(run_dir, True)
            if not success:
                raise PermissionError(error)
            # 文件安全校验
            success, error = GraphUtils.safe_check_load_file_path(file_path)
            if not success:
                raise PermissionError(error)
            # 读取文件比较耗时，支持onlyCheck参数，仅进行安全校验
            if only_check:
                return True, None
            # 尝试解析 JSON 文件,校验文件内容是否合理
            with _thread_local_lock:
                with open(file_path, 'r', encoding='utf-8') as f:
                    return json.load(f), None
        except json.JSONDecodeError:
            logger.error(f'Error: File "{file_path}" is not a valid JSON file!')
            return None, "File is not a valid JSON file!"
        except Exception as e:
            logger.error(f'Error: File "{file_path}" is not accessible. Error: {e}')
            return None, e

    @staticmethod
    def safe_check_save_file_path(file_path, is_dir=False):
        file_path = os.path.normpath(file_path)  # 标准化路径
        real_path = os.path.realpath(file_path)
        safe_base_dir = GraphState.get_global_value('logdir')
        try:
            # 安全验证：路径长度检查
            if len(file_path) > FILE_PATH_MAX_LENGTH:
                raise PermissionError(f"Path length exceeds limit")
            if not is_dir and not os.path.exists(file_path):
                return True, None
            st = os.stat(file_path)
            # 安全验证：禁止符号链接文件
            if os.path.islink(file_path):
                raise PermissionError("The target file is a symbolic link")            
            # 安全验证：检查目录是否存在，如果不存在则创建
            if is_dir and not os.path.exists(real_path):
                os.makedirs(real_path, exist_ok=True)
                os.chmod(file_path, 0o640)
            # 权限校验：检查是否有写权限
            if not os.stat(file_path).st_mode & stat.S_IWUSR:
                raise PermissionError(f"No write permission for directory\n")
            # 安全验证： 非windows系统下，属主检查
            if os.name != 'nt':
                current_uid = os.getuid() 
                # 如果是root用户，跳过后续权限检查
                if current_uid == 0:
                    logger.warning("""Security Warning: Do not run this tool as root. 
                                   Running with elevated privileges may compromise system security. 
                                   Use a regular user account.""")
                    return True, None
                # 属主检查
                if st.st_uid != current_uid:
                    raise PermissionError(f"Directory is not owned by the current user")
                # group和其他用户不可写检查
                if st.st_mode & PERM_GROUP_WRITE or st.st_mode & PERM_OTHER_WRITE:
                    raise PermissionError(
                        f"Directory has group or other write permission, there may be a risk of data tampering.")
            return True, None
        except Exception as e:
            logger.error(e)
            return False, e
    
    @staticmethod
    def safe_check_load_file_path(file_path, is_dir=False):
        # 权限常量定义
        file_path = os.path.normpath(file_path)  # 标准化路径
        real_path = os.path.realpath(file_path)
        st = os.stat(real_path)
        try:
            # 安全验证：路径长度检查
            if len(real_path) > FILE_PATH_MAX_LENGTH:
                raise PermissionError(
                    f"Path is too long (max {FILE_PATH_MAX_LENGTH} characters). Please use a shorter path."
                )
            # 安全检查：文件存在性验证
            if not os.path.exists(real_path):
                raise FileNotFoundError(f"File or directory does not exist,please check the path and ensure it exists.")
            # 安全验证：禁止符号链接文件
            if os.path.islink(file_path):
                raise PermissionError(f"Symbolic links are not allowed,Use a real file path instead.")
            # 安全验证：文件类型检查（防御TOCTOU攻击）
            # 文件类型
            if not is_dir and not os.path.isfile(real_path):
                raise PermissionError(
                    f"Path is not a regular file."
                     "make sure the path points to a valid file (not a directory or device)."
                )
            # 目录类型
            if is_dir and not Path(real_path).is_dir():
                raise PermissionError(
                    f"Expected a directory, but it does not exist or is not a directory."
                     "Please check the path and ensure it is a valid directory."
                )
            # 可读性检查
            if not st.st_mode & stat.S_IRUSR:
                raise PermissionError(
                    f"Current user lacks read permission on file or directory"
                     "Run 'chmod u+r \"<path>\"' to grant read access"
                )
            # 文件大小校验
            if not is_dir and os.path.getsize(file_path) > MAX_FILE_SIZE:
                file_size = GraphUtils.bytes_to_human_readable(os.path.getsize(file_path))
                max_size = GraphUtils.bytes_to_human_readable(MAX_FILE_SIZE)
                raise PermissionError(
                    f"File size exceeds limit ({file_size} > {max_size})."
                     "reduce file size or adjust MAX_FILE_SIZE if needed"
                )
            # 非windows系统下，属主检查
            if os.name != 'nt':
                current_uid = os.getuid() 
                # 如果是root用户，跳过后续权限检查
                if current_uid == 0:
                    logger.warning("""Security Warning: Do not run this tool as root. 
                                   Running with elevated privileges may compromise system security. 
                                   Use a regular user account.""")
                    return True, None 
                # 属主检查
                if st.st_uid != current_uid:
                    raise PermissionError(
                        f"File or directory is not owned by current user,"
                         "Run 'chown <user> \"<path>\"' to fix ownership."
                    )
                # group和其他用户不可写检查
                if st.st_mode & PERM_GROUP_WRITE or st.st_mode & PERM_OTHER_WRITE:
                    raise PermissionError(
                        f"File has insecure permissions: group or others have write access. "
                         "Run 'chmod go-w \"<path>\"' to remove write permissions for group and others."
                    )
            return True, None
        except Exception as e:
            logger.error(e)
            return False, e
       
    @staticmethod
    def find_config_files(run_name):
        """
        在指定目录下查找所有以 .vis.config 结尾的文件（不包括子目录）
        :param directory: 要搜索的目标目录路径
        :return: 包含所有匹配文件路径的列表
        """
        runs = GraphState.get_global_value('runs', {})
        run = runs.get(run_name)
        dir_path = Path(run)
        try:
            if GraphUtils.safe_check_load_file_path(run, True):
                return [
                    file.name for file in dir_path.iterdir()
                    if file.is_file() and file.name.endswith('.vis.config')
                ]
            else:
                return []
        except Exception as e:
            logger.error(e)
            return []

    @staticmethod
    def compare_tag_names(a: str, b: str) -> int:
        """自然排序比较函数，支持路径格式和数字顺序"""

        # 辅助函数：标准化路径并分割组件
        def split_components(s: str) -> list:
            s = s.replace('\\', '/')  # 统一路径分隔符
            return [c for c in re.split(r'[/_]+', s) if c]  # 按 /_ 分割非空组件

        # 辅助函数：将组件拆分为数字/字符串混合的 tokens
        def tokenize(component: str) -> list:
            tokens = []
            buffer = []
            is_num = False

            for char in component:
                if char.isdigit() == is_num and buffer:
                    buffer.append(char)
                else:
                    if buffer:
                        tokens.append(int(''.join(buffer)) if is_num else ''.join(buffer))
                    buffer = [char]
                    is_num = char.isdigit()

            if buffer:
                tokens.append(int(''.join(buffer)) if is_num else ''.join(buffer))
            return tokens

        # 逐级比较组件
        a_comps = split_components(a)
        b_comps = split_components(b)

        for a_part, b_part in zip(a_comps, b_comps):
            a_tokens = tokenize(a_part)
            b_tokens = tokenize(b_part)

            # 比较 token 序列
            for a_tok, b_tok in zip(a_tokens, b_tokens):
                if isinstance(a_tok, int) and isinstance(b_tok, int):
                    if a_tok != b_tok:
                        return a_tok - b_tok
                elif isinstance(a_tok, int):
                    return -1  # 数字优先于字母
                elif isinstance(b_tok, int):
                    return 1
                elif a_tok != b_tok:
                    return -1 if a_tok < b_tok else 1

            # 处理子序列长度差异
            if len(a_tokens) != len(b_tokens):
                return len(a_tokens) - len(b_tokens)

        # 处理组件数量差异
        return len(a_comps) - len(b_comps)

    @staticmethod
    def sort_data(data: dict) -> dict:
        """自然排序比较函数，支持路径格式和数字顺序"""
        sorted_data = {}
        sorted_keys = sorted(data.keys(), key=cmp_to_key(GraphUtils.compare_tag_names))
        for k in sorted_keys:
            # 对每个键对应的值列表进行排序
            sorted_values = sorted(data.get(k, {}).get('tags'), key=cmp_to_key(GraphUtils.compare_tag_names))
            sorted_data[k] = {'type': data.get(k, {}).get('type'), 'tags': sorted_values}

        return sorted_data
            
    @staticmethod
    def is_safe_string(s: str) -> bool:
        """
        安全检查字符串是否包含恶意内容（XSS 防护）
        """
        if not isinstance(s, str):
            return False

        # 1. 长度限制（字符数，非字节）
        if len(s) > FILE_PATH_MAX_LENGTH:
            return False

        # 2. 转换为小写用于检测
        s_lower = s.lower().strip()

        # 3. 黑名单关键词（常见 XSS 向量）
        dangerous_patterns = [
            '<script', '<img', '<svg', '<iframe', '<video', '<audio',
            'javascript:', 'vbscript:', 'data:text/html',
            'onload=', 'onerror=', 'onmouseover=', 'onclick=',
            'eval(', 'alert(', 'document.cookie', 'document.location',
            'window.location', 'innerHTML', 'outerHTML', 'document.write'
        ]

        if any(pattern in s_lower for pattern in dangerous_patterns):
            return False

        return True
    
    @staticmethod
    def escape_html(input_str):
        """
        将字符串中的特殊 HTML 字符转义为 HTML 实体。
        """
        html_escape_map = {
            '&': '&amp;',
            '<': '&lt;',
            '>': '&gt;',
            '"': '&quot;',
            "'": '&#39;',
            '/': '&#x2F;',
        }
        # 使用 str.translate 配合 str.maketrans 提高性能
        translate_table = str.maketrans(html_escape_map)
        return input_str.translate(translate_table)

    @staticmethod
    def validate_colors_param(colors_json: str):
        # 合法颜色正则：#FFFFFF 格式，不区分大小写
    
        """
        校验 colors 参数
        返回: (是否合法, 错误信息, 解析后的数据)
        """

        if not isinstance(colors_json, dict):
            return False, GraphUtils.t('colorsNotObject'), {}

        if len(colors_json) == 0:
            return False, GraphUtils.t('colotsEmpty'), {}

        for key, value in colors_json.items():
            # 2. 校验颜色键
            if not re.match(COLOR_PATTERN, key):
                return False, f"{GraphUtils.t('illegalColorKey')}{key}", {}

            if not isinstance(value, dict):
                return False, f"{GraphUtils.t('colorValuesNotObject')}{key}", {}

            if 'value' not in value:
                return False, f"{GraphUtils.t('missingValueField')}{key}", {}

            # 3. 校验 value 字段
            val = value['value']
            if isinstance(val, list):
                if len(val) != 2:
                    return False, f"{GraphUtils.t('notArrayOfLength2')}{key}", {}
                if not all(isinstance(x, (int, float)) for x in val):
                    return False, f"{GraphUtils.t('notArrayConsistNumbers')}{key}", {}
                if val[0] >= val[1]:
                    return False, f"{GraphUtils.t('invalidValueRange')}{key}", {}
            elif isinstance(val, str):
                if val not in ["无匹配节点", "N/A", "No matching nodes"]:
                    return False, f"{GraphUtils.t('unsupportedValue')}{val}", {}
            else:
                return False, f"{GraphUtils.t('valueTypeError')}{key}", {}

            # 4. 校验 description
            desc = value.get('description', '')
            if not GraphUtils.is_safe_string(desc):
                return False, f"{GraphUtils.t('descriptionError')}{key}", {}
            value['description'] = GraphUtils.escape_html(desc)

        return True, None, colors_json



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import json
import time
import sqlite3
from tensorboard.util import tb_logging

from .graph_repo_base import GraphRepo
from ..utils.graph_utils import GraphUtils
from ..utils.global_state import GraphState
from ..utils.constant import SINGLE, NPU, BENCH, DataType

logger = tb_logging.get_logger()
DB_TYPE = DataType.DB.value


class GraphRepoDB(GraphRepo):

    def __init__(self, db_path):
        self.db_path = db_path
        self.repo_type = DB_TYPE
        self._initialize_db_connection()           

    def get_db_connection(self):
        return self.conn

    # DB: 查询配置表信息
    def query_config_info(self):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        query = f"SELECT * FROM tb_config"
        try:
            with conn as c:
                cursor = c.execute(query)
                rows = cursor.fetchall()
                
            record = dict(rows[0])
            # 构建最终的 data 对象
            config_info = {
                "microSteps": record.get('micro_steps', 1) or 1,
                "tooltips": GraphUtils.safe_json_loads(record.get('tool_tip')),
                "overflowCheck": bool(record.get('overflow_check', 1)),
                "isSingleGraph": not record.get('graph_type') == 'compare',
                "colors": GraphUtils.safe_json_loads(record.get('node_colors')),
                "matchedConfigFiles": [],
                "task": record.get('task', ''),
                "ranks": GraphUtils.safe_json_loads(record.get('rank_list')),
                "steps": GraphUtils.safe_json_loads(record.get('step_list')),
            }
            return config_info
        except Exception as e:
            logger.error(f"Failed to query config info: {e}")
            return {}
        finally:
            conn.close()

    # DB：查询根节点信息
    def query_root_nodes(self, graph_type, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        graph_type = graph_type if graph_type != SINGLE else NPU
        query = """
        SELECT 
            node_name,
            up_node,
            sub_nodes,
            node_type,
            matched_node_link,
            precision_index,
            overflow_level,
            matched_distributed 
        FROM 
            tb_nodes 
        WHERE   
            step = ?
            AND rank = ? 
            AND data_source = ? 
            AND up_node = '' 
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, graph_type))
                rows = cursor.fetchall()
            if len(rows) > 0:
                return self._convert_db_to_object(dict(rows[0]))
            else:
                return {}
        except Exception as e:
            logger.error(f"Failed to query root nodes: {e}")
            return {}
        finally:
            conn.close()
    
    # DB：查询当前节点的所有父节点信息
    def query_up_nodes(self, node_name, graph_type, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        graph_type = graph_type if graph_type != SINGLE else NPU
        # 现根据节点名称查询节点信息，根据up_node字段得到父节点名称
        # 再根据父节点名称查询父节点信息
        # 递归查询父节点，直到根节点
        query = """
            WITH RECURSIVE parent_chain AS (
                SELECT child.id, child.node_name, child.up_node, child.data_source, child.rank, child.step, 0 AS level
                FROM 
                    tb_nodes child
                WHERE  
                    child.step = ?
                    AND child.rank = ?
                    AND child.data_source = ?
                    AND child.node_name = ?

                UNION ALL

                SELECT 
                    parent.id, 
                    parent.node_name, 
                    parent.up_node,
                    parent.data_source, 
                    parent.rank, 
                    parent.step, 
                    pc.level + 1
                FROM 
                    tb_nodes parent
                INNER JOIN parent_chain pc 
                    ON parent.data_source = pc.data_source
                    AND parent.node_name  = pc.up_node
                    AND parent.rank = pc.rank
                    AND parent.step = pc.step
                WHERE 
                    pc.up_node IS NOT NULL 
                    AND pc.up_node != ''
                )
            SELECT 
                tb_nodes.id,
                tb_nodes.data_source,
                tb_nodes.node_name,
                tb_nodes.up_node,
                tb_nodes.sub_nodes,
                tb_nodes.node_type,
                tb_nodes.matched_node_link,
                tb_nodes.precision_index,
                tb_nodes.overflow_level,
                tb_nodes.matched_distributed
            FROM 
                tb_nodes
            WHERE 
                id IN (SELECT id FROM parent_chain)
            ORDER BY (
                SELECT 
                    level 
                FROM 
                    parent_chain pc 
                WHERE 
                    pc.node_name = tb_nodes.node_name) 
                ASC
        """ 
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, graph_type, node_name))
                rows = cursor.fetchall()
            up_nodes = {}
            for row in rows:
                dict_row = self._convert_db_to_object(dict(row))
                up_nodes[row['node_name']] = dict_row
            return up_nodes
        except Exception as e:
            logger.error(f"Failed to query up nodes: {e}")
            return {}
        finally:
            conn.close()

    # DB: 查询待匹配节点的信息，构造graph data
    def query_matched_nodes_info(self, npu_node_name, bench_node_name, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        query = """
            SELECT 
                id,
                node_name,
                node_type,
                up_node,
                sub_nodes,
                data_source,
                input_data,
                output_data,
                matched_node_link
            FROM tb_nodes
            WHERE step = ? AND rank = ? AND data_source = ? AND node_name = ?
            """
        npu_nodes = {}
        bench_nodes = {}
        opposite_npu_node_name = GraphUtils.get_opposite_node_name(npu_node_name)
        opposite_bench_node_name = GraphUtils.get_opposite_node_name(bench_node_name)
        # 定义查询参数列表：(graph_type, node_name, target_dict_key)
        queries = [
            (NPU, npu_node_name, 'npu'),
            (NPU, opposite_npu_node_name, 'npu_opposite'),
            (BENCH, bench_node_name, 'bench'),
            (BENCH, opposite_bench_node_name, 'bench_opposite'),
        ]
        # 存储结果的字典
        nodes_dict = {}
        try:
            with conn as c:
                for graph_type, node_name, key in queries:
                    if not node_name:  # 可选：跳过空 node_name
                        continue
                    cursor = c.execute(query, (step, rank, graph_type, node_name))
                    rows = cursor.fetchall()
                    if rows:
                        node_obj = self._convert_db_to_object(dict(rows[0]))
                        nodes_dict[key] = {node_obj.get('node_name'): node_obj}
                    else:
                        nodes_dict[key] = {}

            npu_nodes = nodes_dict.get('npu', {}) | nodes_dict.get('npu_opposite', {})
            bench_nodes = nodes_dict.get('bench', {}) | nodes_dict.get('bench_opposite', {})    
            result = self._convert_to_graph_json(npu_nodes, bench_nodes)
            return result
        except Exception as e:
            logger.error(f"Failed to query matched nodes info: {e}")
            return self._convert_to_graph_json({}, {})
        finally:
            conn.close()
            
    # DB: 查询待匹配节点及其子节点的信息，递归查询当前节点信息和其所有的子节点信息，一直叶子节点
    def query_node_and_sub_nodes(self, npu_node_name, bench_node_name, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        query = """
            WITH RECURSIVE descendants AS (
            -- 初始节点选择
            SELECT 
                id,
                node_name,
                node_type,
                up_node,
                sub_nodes,
                data_source,
                input_data,
                output_data,
                matched_node_link,
                node_order,
                step,
                rank
            FROM tb_nodes
            WHERE step = ? AND rank = ? AND data_source = ? AND node_name = ?

            UNION ALL

            -- 递归部分
            SELECT 
                child.id,
                child.node_name,
                child.node_type,
                child.up_node,
                child.sub_nodes,
                child.data_source,
                child.input_data,
                child.output_data,
                child.matched_node_link,
                child.node_order,
                child.step,
                child.rank
            FROM descendants d
            JOIN json_each(d.sub_nodes) AS je          -- 将 sub_nodes JSON 数组展开为多行
            JOIN tb_nodes child 
                ON child.node_name = je.value         -- 子节点名称匹配
                AND child.step = d.step
                AND child.rank = d.rank
                AND child.data_source = d.data_source
            WHERE 
                d.sub_nodes IS NOT NULL               -- 父节点的 sub_nodes 不为 NULL
                AND d.sub_nodes != ''               -- 不是空
                AND d.sub_nodes != '[]' 
                AND json_type(d.sub_nodes) = 'array'  -- 确保是合法 JSON 数组
        )
        SELECT * FROM descendants
        """ 
        npu_nodes = {}
        bench_nodes = {}
        opposite_npu_node_name = GraphUtils.get_opposite_node_name(npu_node_name)
        opposite_bench_node_name = GraphUtils.get_opposite_node_name(bench_node_name)
        # 定义查询参数列表：(graph_type, node_name, target_dict_key)
        queries = [
            (NPU, npu_node_name, 'npu'),
            (NPU, opposite_npu_node_name, 'npu_opposite'),
            (BENCH, bench_node_name, 'bench'),
            (BENCH, opposite_bench_node_name, 'bench_opposite'),
        ]
        # 存储结果的字典
        nodes_dict = {}
        try:
            with conn as c:
                for graph_type, node_name, key in queries:
                    if not node_name:  # 可选：跳过空 node_name
                        continue
                    cursor = c.execute(query, (step, rank, graph_type, node_name))
                    nodes_dict[key] = self._fetch_and_convert_rows(cursor)
            npu_nodes = nodes_dict.get('npu', {}) | nodes_dict.get('npu_opposite', {})
            bench_nodes = nodes_dict.get('bench', {}) | nodes_dict.get('bench_opposite', {})    
            result = self._convert_to_graph_json(npu_nodes, bench_nodes)
            return result
        except Exception as e:
            logger.error(f"Failed to query node and sub nodes: {e}")
            return {'NPU': {}, 'Bench': {}}
        finally:
            conn.close()
    
    # DB：查询配置文件中的待匹配节点信息
    def query_matched_nodes_info_by_config(self, match_node_links, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        query = """
            SELECT 
                id,
                node_name,
                node_type,
                up_node,
                sub_nodes,
                data_source,
                input_data,
                output_data,
                matched_node_link 
            FROM 
                tb_nodes 
            WHERE 
                step = ?
                AND rank = ?
                AND data_source = ?
                AND node_name IN ({}) 
            """.format(','.join(['?'] * len(match_node_links)))
        try:
            with conn as c:
                npu_node_names = list(match_node_links.keys())
                bench_node_names = list(match_node_links.values())
                npu_cursor = c.execute(query, (step, rank, NPU, *npu_node_names))
                bench_cursor = c.execute(query, (step, rank, BENCH, *bench_node_names))
                npu_nodes = self._fetch_and_convert_rows(npu_cursor)
                bench_nodes = self._fetch_and_convert_rows(bench_cursor)
            result = self._convert_to_graph_json(npu_nodes, bench_nodes)
            return result
        except Exception as e:
            logger.error(f"Failed to query nodes info: {e}")
            return {}
        finally:
            conn.close()
    
    # DB: 查询所有以当前为父节点的子节点
    def query_sub_nodes(self, node_name, graph_type, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        graph_type = graph_type if graph_type != SINGLE else NPU
        query = """
            SELECT 
                node_name,
                up_node,
                sub_nodes,
                node_type,
                micro_step_id,
                matched_node_link,
                precision_index,
                overflow_level,
                matched_distributed 
            FROM 
                tb_nodes 
            WHERE  
                step = ?
                AND rank = ?
                AND data_source = ? 
                AND up_node = ?
            ORDER BY
                node_order ASC
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, graph_type, node_name))
                rows = cursor.fetchall()
            sub_nodes = {}
            for row in rows:
                dict_row = self._convert_db_to_object(dict(row))
                sub_nodes[row['node_name']] = dict_row
            return sub_nodes
        except Exception as e:
            logger.error(f"Failed to query sub nodes: {e}")
            return {}
        finally:
            conn.close()

    # DB: 查询当前节点信息
    def query_node_info(self, node_name, graph_type, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        graph_type = graph_type if graph_type != SINGLE else NPU
        query = """
            SELECT 
                n.*,
                d.stack_info 
            FROM 
                tb_nodes n
            JOIN tb_stack d ON n.stack_id = d.id
            WHERE 
                n.step = ?
                AND n.rank = ? 
                AND n.data_source = ? 
                AND n.node_name = ?
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, graph_type, node_name))
                rows = cursor.fetchall()
            if len(rows) > 0:
                return self._convert_db_to_object(dict(rows[0]))
            else:
                return {}
        except Exception as e:
            logger.error(f"Failed to query node info: {e}")
            return {}
        finally:
            conn.close()
    
    # DB: 查询单图节点名称列表
    def query_node_name_list(self, rank, step, micro_step):
        conn = self._initialize_db_connection()
        if not conn:
            return []
        query = """
            SELECT 
                node_name
            FROM 
                tb_nodes 
            WHERE 
                step = ?
                AND rank = ?
                AND (? = -1 OR micro_step_id = ?)
                AND data_source = 'NPU'
            ORDER BY
                node_order ASC
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, micro_step, micro_step))
                rows = cursor.fetchall()
            return [row['node_name'] for row in rows]
        except Exception as e:
            logger.error(f"Failed to query node name list: {e}")
            return []
        finally:
            conn.close()

    # DB: 查询已匹配节点列表，未匹配节点列表，所有的节点列表
    def query_all_node_info_in_one(self, rank, step, micro_step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        try:
            # 查找缓存
            all_node_info_cache = GraphState.get_global_value('all_node_info_cache', {})
            cache = f'{rank}_{step}_{micro_step}'
            if all_node_info_cache.get(cache) is not None:
                return all_node_info_cache.get(cache) 
            # 查询数据库
            # 单次查询：获取 node_name 和 matched_node_link
            query = """
                SELECT 
                    node_name,
                    data_source,
                    matched_node_link 
                FROM 
                    tb_nodes 
                WHERE 
                    step = ?
                    AND rank = ? 
                    AND (? = -1 OR micro_step_id = ?)
                ORDER BY
                    node_order ASC
            """
            
            with conn as conn:
                cursor = conn.execute(query, (step, rank, micro_step, micro_step))
                rows = cursor.fetchall()
            # 初始化结果
            npu_node_list = []
            bench_node_list = []
            npu_match_node = {}  # {node_name: last_matched_link}
            bench_match_node = {}
            npu_unmatch_node = []
            bench_unmatch_node = []

            # 一次性遍历结果，分类处理
            for row in rows:
                node_name = row['node_name']
                matched_link_str = row['matched_node_link']
                if row['data_source'] == NPU:
                    npu_node_list.append(node_name)
                    # 解析 matched_node_link
                    matched_link = GraphUtils.safe_json_loads(matched_link_str)
                    # 判断是否为有效匹配（非空列表）
                    if isinstance(matched_link, list) and len(matched_link) > 0:
                        npu_match_node[node_name] = matched_link[-1]  # 取最后一个匹配项
                    else:
                        npu_unmatch_node.append(node_name)
                elif row['data_source'] == BENCH:
                    bench_node_list.append(node_name)
                    # 解析 matched_node_link
                    matched_link = GraphUtils.safe_json_loads(matched_link_str)
                    # 判断是否为有效匹配（非空列表）
                    if isinstance(matched_link, list) and len(matched_link) > 0:
                        bench_match_node[node_name] = matched_link[-1]  # 取最后一个匹配项
                    else:
                        bench_unmatch_node.append(node_name)
                else:
                    logger.error(f"Invalid data source: {row['data_source']}")
            all_node_info = {
                'npu_node_list': npu_node_list,
                'bench_node_list': bench_node_list,
                'npu_match_node': npu_match_node,
                'bench_match_node': bench_match_node,
                'npu_unmatch_node': npu_unmatch_node,
                'bench_unmatch_node': bench_unmatch_node
            }
            all_node_info_cache = GraphState.get_global_value('all_node_info_cache', {})
            all_node_info_cache[cache] = all_node_info
            return all_node_info

        except Exception as e:
            logger.error(f"Failed to query all node info: {e}")
            return {
                'npu_node_list': [],
                'bench_node_list': [],
                'npu_match_node': {},
                'bench_match_node': {},
                'npu_unmatch_node': [],
                'bench_unmatch_node': []
            }
        finally:
            conn.close()

    # # DB：根据step rank modify match_node_link查询已经修改的匹配成功的节点关系
    def query_modify_matched_nodes_list(self, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        query = """
            SELECT 
                node_name,
                matched_node_link 
            FROM 
                tb_nodes 
            WHERE 
                step = ?
                AND rank = ? 
                AND modified = 1
                AND matched_node_link IS NOT NULL
                AND matched_node_link != '[]'
                AND matched_node_link != ''
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank))
                rows = cursor.fetchall()
            result = {}
            for row in rows:
                matched_node_link = GraphUtils.safe_json_loads(row['matched_node_link'])
                node_name = row['node_name']
                if isinstance(matched_node_link, list) and len(matched_node_link) > 0:
                    result[node_name] = matched_node_link[-1]  # 取最后一个匹配项
            return result
        except Exception as e:
            logger.error(f"Failed to query modify matched nodes list: {e}")
            return {}
        finally:
            conn.close()
            
    # DB: 根据精度误差查询节点信息
    def query_node_list_by_precision(self, step, rank, micro_step, values, is_filter_unmatch_nodes):
        conn = self._initialize_db_connection()
        if not conn:
            return []
        # 准备占位符
        conditions = []
        placeholders = []
        params = []
        conditions.append("step = ?")
        conditions.append("rank = ?")
        conditions.append("data_source = 'NPU'")
        conditions.append("(? = -1 OR micro_step_id = ?)")
        conditions.append("(sub_nodes = '' OR sub_nodes IS NULL OR sub_nodes = '[]')") 
        for value in values:
            placeholder = "(precision_index BETWEEN ? AND ?)"
            placeholders.append(placeholder)
            params.extend(value)
    
        if is_filter_unmatch_nodes:
            placeholders.append("(matched_node_link = '' OR matched_node_link IS NULL OR matched_node_link = '[]')")

        if len(placeholders) > 0:
            conditions.append(f"({'OR'.join(placeholders)})")
        query = f"""
            SELECT 
                node_name
            FROM
                tb_nodes 
            WHERE 
                {" AND ".join(conditions)}
            ORDER BY
                node_order ASC
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, micro_step, micro_step, *params))
                rows = cursor.fetchall()
            node_list = [row['node_name'] for row in rows]
            return node_list
        except Exception as e:
            logger.error(f"Failed to query node list by precision: {e}")
            return []
        finally:
            conn.close()

    # DB: 根据溢出查询节点信息
    def query_node_list_by_overflow(self, step, rank, micro_step, values):
        conn = self._initialize_db_connection()
        if not conn:
            return []
        # 准备占位符
        conditions = []
        conditions.append("step = ?")
        conditions.append("rank = ?")
        conditions.append("data_source = 'NPU'")
        conditions.append("(? = -1 OR micro_step_id = ?)")
        placeholders = ", ".join(["?"] * len(values)) 
        query = f"""
            SELECT 
                node_name
            FROM
                tb_nodes 
            WHERE 
                step = ?
                AND rank = ? 
                AND data_source = 'NPU'
                AND (? = -1 OR micro_step_id = ?)
                AND (sub_nodes = '' OR sub_nodes IS NULL OR sub_nodes = '[]')
                AND overflow_level IN ({placeholders})
            ORDER BY
                node_order ASC
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, micro_step, micro_step, *values))
                rows = cursor.fetchall()
            node_list = [row['node_name'] for row in rows]
            return node_list
        except Exception as e:
            logger.error(f"Failed to query node list by overflow: {e}")
            return []
        finally:
            conn.close()

    # DB：查询节点信息
    def query_node_info_by_data_source(self, step, rank, data_source):
        conn = self._initialize_db_connection()
        if not conn:
            return {}
        query = """
            SELECT 
                node_name, 
                matched_node_link,
                output_data,
                precision_index,
                sub_nodes
            FROM 
                tb_nodes 
            WHERE 
                step = ?
                AND rank = ? 
                AND data_source = ?
        """
        try:
            with conn as c:
                cursor = c.execute(query, (step, rank, data_source))
                nodes = self._fetch_and_convert_rows(cursor)
            return nodes
        except Exception as e:
            logger.error(f"Failed to query node info: {e}")
            return {}
        finally:
            conn.close()

    # DB：更新config的colors
    def update_config_colors(self, colors):
        conn = self._initialize_db_connection()
        if not conn:
            return False
        query = """
            UPDATE 
                tb_config 
            SET 
                node_colors = ?
            WHERE
                id=1
        """
        try:
            with conn as c:
                c.execute(query, (json.dumps(colors),))
            return True
        except Exception as e:
            logger.error(f"Failed to update config colors: {e}")
            return False
        finally:
            conn.close()

    # DB：批量更新节点信息
    def update_nodes_info(self, nodes_info, rank, step):
        conn = self._initialize_db_connection()
        if not conn:
            return False
        # 取消匹配和匹配都要走这个逻辑        
        try:
            data = [
                (
                    json.dumps(node['matched_node_link']),
                    json.dumps(node['input_data']),
                    json.dumps(node['output_data']),
                    node['precision_index'],
                    step,
                    rank,
                    node['graph_type'],
                    node['node_name']  # WHERE 条件
                )
                for node in nodes_info
            ]
            query = """
                UPDATE tb_nodes 
                SET 
                    matched_node_link = ?,
                    input_data = ?,
                    output_data = ?,
                    precision_index = ?,
                    modified= 1
                WHERE 
                    step = ?
                    AND rank = ? 
                    AND data_source = ? 
                    AND node_name = ?
            """
            with conn as c:
                c.executemany(query, data)
            return True
        except Exception as e:
            logger.error(f"Failed to update nodes info: {e}")
            return False
    
    def update_nodes_precision_error(self, update_data):
        conn = self._initialize_db_connection()
        if not conn:
            return False
        query = """
            UPDATE 
                tb_nodes
            SET
                precision_index = ?
            WHERE
                step = ?
                AND rank = ?
                AND data_source = 'NPU'
                AND node_name = ?
        """
        try:
            conn.executemany(query, update_data)
            conn.commit()
            return True
        except Exception as e:
            logger.error(f"Failed to update precision error: {e}")
            return False
        finally:
            conn.close()

    def _initialize_db_connection(self):
        try:
            # 目录安全校验
            dir_path = str(os.path.dirname(self.db_path))
            success, error = GraphUtils.safe_check_load_file_path(dir_path, True)
            if not success:
                raise PermissionError(error)
            # 文件安全校验
            success, error = GraphUtils.safe_check_load_file_path(self.db_path)
            if not success:
                raise PermissionError(error)
            conn = sqlite3.connect(self.db_path, check_same_thread=False)
            conn.row_factory = sqlite3.Row
            # 提升性能的 PRAGMA 设置
            conn.execute("PRAGMA journal_mode = WAL;")
            conn.execute("PRAGMA synchronous = NORMAL;")  # 或 OFF（不安全）
            conn.execute("PRAGMA cache_size = 40000;")
            conn.execute("PRAGMA wal_autocheckpoint = 0;")
            self.is_db_connected = conn is not None
            self.conn = conn
            return conn
        
        except Exception as e:
            logger.error(f"Failed to connect to database: {e}")
            if 'conn' in locals() and conn:
                try:
                    conn.close()
                except Exception as close_error:
                    logger.error(f"Failed to close database connection: {close_error}")
            self.conn = None
            return None
    
    def _fetch_and_convert_rows(self, cursor):
        """
        Helper function to fetch rows from cursor and convert them.
        :param cursor: SQLite cursor object
        :return: Dictionary of nodes keyed by node_name
        """
        nodes = {}
        for row in cursor.fetchall():
            dict_row = self._convert_db_to_object(dict(row))
            nodes[row['node_name']] = dict_row
        return nodes
    
    def _convert_to_graph_json(self, npu_nodes, bench_nodes):
        graph_data = {
            "NPU": {
                "node": npu_nodes,
            },
            "Bench": {
                "node": bench_nodes,
            }
        }
        return graph_data

    def _convert_db_to_object(self, data):
        object_res = {
            "id": data.get('node_name'),
            "node_name": data.get('node_name'),
            "node_type": int(data.get('node_type')) if data.get('node_type') is not None else 0,
            "output_data": GraphUtils.safe_json_loads(data.get('output_data') or "{}"),
            "input_data": GraphUtils.safe_json_loads(data.get('input_data') or "{}"),
            "upnode": data.get('up_node'),
            "subnodes": GraphUtils.safe_json_loads(data.get('sub_nodes') or "[]"),
            "matched_node_link": GraphUtils.safe_json_loads(data.get('matched_node_link') or "[]"),
            "stack_info": GraphUtils.safe_json_loads(data.get('stack_info') or "[]"),
            "micro_step_id": int(data.get('micro_step_id')) if data.get('micro_step_id') is not None else -1,
            "data": {
                "precision_index": data.get('precision_index'),
                'overflow_level': data.get('overflow_level'),
            },
            "parallel_merge_info": GraphUtils.safe_json_loads(data.get('parallel_merge_info') or "[]"),
            "matched_distributed": GraphUtils.safe_json_loads(data.get('matched_distributed') or "[]"),
            "modified": int(data.get('modified')) if data.get('modified') is not None else 0,
        }
        return object_res
    



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.utils import CompareException
from msprobe.core.common.file_utils import create_directory
from msprobe.core.common.exceptions import FileCheckException
from msprobe.mindspore.common.log import logger
from msprobe.core.compare.utils import compare_distributed_inner
from msprobe.mindspore.compare.ms_graph_compare import GraphMSComparator
from msprobe.core.compare.utils import print_compare_ends_info


def ms_compare_distributed(npu_dump_dir, bench_dump_dir, output_path, **kwargs):
    compare_distributed_inner(npu_dump_dir, bench_dump_dir, output_path, **kwargs)


def ms_graph_compare(args):
    try:
        create_directory(args.output_path)
    except (CompareException, FileCheckException) as error:
        logger.error('Compare failed. Please check the arguments and do it again!')
        return
    ms_comparator = GraphMSComparator(args)
    ms_comparator.compare_core()
    print_compare_ends_info()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Any

CHECK_FIELDS = ['Max', 'Min', 'Mean']
OVERFLOW_VALUES = ['inf', '-inf', 'nan']


def has_nan_inf(value: Any) -> bool:
    """检查值是否包含NaN或Inf"""
    if isinstance(value, dict):
        for k, v in value.items():
            if k in CHECK_FIELDS and str(v).lower() in OVERFLOW_VALUES:
                return True
    return False



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
StandardDeviation algorithm. This file mainly involves the compare function.
"""

import numpy as np

from msprobe.msaccucmp.algorithm_manager.algorithm_parameter import AlgorithmParameter
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager


def compare(my_output_dump_data: any, ground_truth_dump_data: any, args: AlgorithmParameter) -> (str, str):
    """
    compare the my output dump data and the ground truth dump data
    by standard deviation
    :param my_output_dump_data: the my output dump data to compare
    :param ground_truth_dump_data: the ground truth dump data to compare
    :param args: the algorithm parameter
    :return: the result of the standard deviation value and error message (the default is "")
    """
    _ = args  # Bypassing parameter is not used
    left_std = np.std(my_output_dump_data, dtype=np.float64)
    right_std = np.std(ground_truth_dump_data, dtype=np.float64)
    left_mean = np.mean(my_output_dump_data, dtype=np.float64)
    right_mean = np.mean(ground_truth_dump_data, dtype=np.float64)
    left_std = 0.0 if abs(left_std) < ConstManager.MINIMUM_VALUE else left_std
    left_mean = 0.0 if abs(left_mean) < ConstManager.MINIMUM_VALUE else left_mean
    right_std = 0.0 if abs(right_std) < ConstManager.MINIMUM_VALUE else right_std
    right_mean = 0.0 if abs(right_mean) < ConstManager.MINIMUM_VALUE else right_mean
    return "(%.3f;%.3f),(%.3f;%.3f)" % (left_mean, left_std, right_mean, right_std), ""



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import defaultdict
import mindspore as ms
from mindspore.ops.primitive import Primitive

from msprobe.core.common.utils import Const
from msprobe.core.dump.service import BaseService
from msprobe.mindspore.dump.cell_processor import CellProcessor
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.common.utils import (
    get_rank_if_initialized,
    is_mindtorch,
    get_cells_and_names_with_index
)
from msprobe.mindspore.dump.dump_processor.hook_cell.api_register import get_api_register, ApiTemplate
from msprobe.mindspore.dump.dump_processor.hook_cell.ms_hook_manager import MindsporeHookManager
from msprobe.mindspore.dump.dump_processor.hook_cell.primitive_hooks import PrimitiveHookService
from msprobe.mindspore.dump.dump_processor.jit_dump import JitDump

try:
    from mindspore.common._pijit_context import PIJitCaptureContext
except ImportError:
    pijit_label = False
else:
    pijit_label = True


class MindsporeService(BaseService):
    @property
    def _get_framework_type(self):
        return Const.MT_FRAMEWORK if is_mindtorch() else Const.MS_FRAMEWORK

    @staticmethod
    def _get_current_rank():
        return get_rank_if_initialized()

    def empty(self, *args, **kwargs):
        pass

    def reset_status(self):
        self._reset_status()

    def _init_specific_components(self):
        self.logger = logger
        self.api_register = get_api_register()
        self.primitive_hook_service = PrimitiveHookService(self)
        self.cell_processor = CellProcessor(self.data_collector.scope)
        self.hook_manager = MindsporeHookManager(self.data_collector, self.config)
        self._setup_jit_context()
        self.api_template = ApiTemplate

    def _setup_jit_context(self):
        if self.config.level in [Const.LEVEL_MIX, Const.LEVEL_L1]:
            JitDump.set_config(self.config)
            JitDump.set_data_collector(self.data_collector)
            if hasattr(ms.common.api, "_MindsporeFunctionExecutor"):
                ms.common.api._MindsporeFunctionExecutor = JitDump
            else:
                ms.common.api._JitExecutor = JitDump
            ms.common.api._PyNativeExecutor.grad = JitDump.grad
            if pijit_label:
                PIJitCaptureContext.__enter__ = self.empty
                PIJitCaptureContext.__exit__ = self.empty

    def _register_module_hook(self):
        self.cell_processor.register_cell_hook(self.model, self.build_hook, self.config)
        self.logger.info_on_rank_0(f"The module {self.config.task} hook function is successfully mounted to the model.")

    def _register_hook(self):
        self._register_primitive_hook()

    def _register_primitive_hook(self):
        if self.config.level not in [Const.LEVEL_MIX, Const.LEVEL_L1]:
            return
        if not self.model or self.config.task not in Const.DUMP_DATA_COLLECTION_LIST:
            return

        primitive_set = set()
        cells_and_names_with_index, _ = get_cells_and_names_with_index(self.model)
        for cells_and_names in cells_and_names_with_index.values():
            for _, cell in cells_and_names:
                for attribute, value in vars(cell).items():
                    if isinstance(value, Primitive):
                        primitive_set.add((attribute, value))

        for pname, primitive in primitive_set:
            primitive_class_name = primitive.__class__.__name__
            primitive_combined_name = pname + Const.SEP + primitive_class_name
            new_primitive = type('NewPrimitive', (primitive.__class__,),
                                    {'__call__': self.primitive_hook_service.wrap_primitive(primitive.__call__,
                                                                                            primitive_combined_name)})
            primitive.__class__ = new_primitive

    def _reset_status(self):
        super()._reset_status()
        self.primitive_hook_service.primitive_counters.clear()
        JitDump.jit_count = defaultdict(int)

    def _change_jit_switch(self, status):
        JitDump.jit_dump_switch = status



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Union
from msprobe.infer.utils.check.checker import Checker, CheckResult, rule


class DictChecker(Checker):

    @rule()
    def is_dict(self) -> Union["DictChecker", CheckResult]:
        return isinstance(self.instance, dict), "Input object is not a dict."

    @rule()
    def is_dict_not_empty(self) -> Union["DictChecker", CheckResult]:
        is_pass = self.is_dict()
        if not is_pass:
            return is_pass
        return bool(self.instance), "Dict is empty."

    @rule()
    def is_key_exists_in_dict(self, key) -> Union["DictChecker", CheckResult]:
        is_pass = self.is_dict()
        if not is_pass:
            return is_pass
        res = self._find_key_in_dict(self.instance, key)
        return res, f"Key='{key}' is not in the dict."

    @rule()
    def is_values_valid(self, default_value=None, **value_rules: Checker) -> Union["DictChecker", CheckResult]:
        is_pass = self.is_dict().passed
        if not is_pass:
            return is_pass, "Input object is not a dict."
        err_msgs = []
        for key, value_rule in value_rules.items():
            value = self.instance.get(key, default_value)
            rule_is_pass = value_rule.check(value)
            if not rule_is_pass:
                is_pass = False
                err_msgs.append(f"{key} is invalid. {str(rule_is_pass)}")

        return is_pass, "\n".join(err_msgs)

    @rule()
    def is_key_type_valid(self, exp_type) -> Union["DictChecker", CheckResult]:
        is_pass = self.is_dict().passed
        if not is_pass:
            return is_pass
        err_msg = []
        for key in self.instance.keys():
            if not isinstance(key, exp_type):
                is_pass = False
                err_msg.append(f'The type of {key} is not {exp_type.__name__}')
        return is_pass, "\n".join(err_msg)

    def _find_key_in_dict(self, my_dict, my_key):
        if not isinstance(my_dict, dict):
            return False
        if my_key in my_dict:
            return True
        for dict_val in my_dict.values():
            if isinstance(dict_val, dict):
                if self._find_key_in_dict(dict_val, my_key):
                    return True
        return False



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch

from msprobe.pytorch.common.log import logger


device = "cpu"
try:
    import torch_npu
    device = "npu"
except ImportError:
    if torch.cuda.is_available():
        device = "cuda"

NAN_TENSOR_ON_DEVICE = None


def get_nan_tensor():
    global NAN_TENSOR_ON_DEVICE
    if not NAN_TENSOR_ON_DEVICE:
        NAN_TENSOR_ON_DEVICE = torch.tensor(torch.nan, device=device)
    return NAN_TENSOR_ON_DEVICE


def get_param_struct(param):
    res = {}
    if isinstance(param, (tuple, list)):
        res['config'] = f'{type(param).__name__}[{len(param)}]'
        for i, x in enumerate(param):
            res[i] = f'size={tuple(x.shape)}, dtype={x.dtype}' if torch.is_tensor(x) else f'{type(x)}'
    elif torch.is_tensor(param):
        res['config'] = 'tensor'
        res['tensor'] = f'size={tuple(param.shape)}, dtype={param.dtype}'
    else:
        res['config'] = f'{type(param)}'
        logger.warning(f'Not support type({type(param)}) now, please check the type of param {param}')
    return res